"""Worker that connects to an LinxProcessor object."""

from pathlib import Path

from PyQt6.QtCore import pyqtSignal as Signal
from PyQt6.QtCore import pyqtSlot as Slot

from pymagnetos.core.gui import BaseWorker
from pymagnetos.linx import LinxProcessor


class Worker(BaseWorker):
    """
    Wrapper around the LinxProcessor class.

    It exposes the LinxProcessor methods as slots and emits a signal when the task is
    finished.

    Additional signals
    ------------------
    sig_export_csv_finished : emits when the results are saved as CSV.
    """

    _type_processor: type[LinxProcessor] = LinxProcessor
    proc: LinxProcessor

    sig_export_csv_finished = Signal()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # Setup methods arguments
        self._after_load_analysis = True

    @Slot()
    def auto_phase(self) -> None:
        """Find the phase offset that maximizes the in-phase signal."""
        self.proc.find_phase_offset()
        self.run_analysis()

    @Slot(str, int, bool)
    def export_csv(self, fname: str | Path, downsample: int, average: bool) -> None:
        """Export demodulated signal as CSV."""
        self.proc.export_csv(fname, downsample=downsample, average=average)
        self.sig_export_csv_finished.emit()



"""Classes to store data as NeXus objects."""

import importlib.metadata
from typing import Any

import nexusformat.nexus as nx


class BaseData(nx.NXgroup):
    """Base Data class integrating NXgroup."""

    def __init__(self, attr: dict = {}, **kwargs) -> None:
        """
        Base Data class integrating NXgroup [1].

        `DataBase` should be used for subclassing. It allows to create a NXgroup with
        attributes specified as a dictionary.

        [1] https://nexpy.github.io/nexpy/treeapi.html#nexusformat.nexus.tree.NXgroup

        Parameters
        ----------
        attr : dict, optional
            Attributes for the NXgroup. Default is an empty dictionary.
        **kwargs : passed to `nexusformat.NXgroup()`.
        """
        super().__init__(**kwargs)

        # Set attributes
        for k, v in attr.items():
            self.attrs[k] = v


class DataRaw(BaseData):
    """Store raw data in a NeXus NXdata group."""

    def __init__(self, attr: dict = {}, **kwargs) -> None:
        """
        Store raw data in a NeXus group.

        The instantiated object will be of type `nexusformat.nexus.NXdata` [1].

        [1] https://nexpy.github.io/nexpy/treeapi.html#nexusformat.nexus.tree.NXdata

        Parameters
        ----------
        attr : dict, optional
            Attribute for the NXdata group. Default is an empty dictionary.
        **kwargs : passed to `pyuson.data.DataBase()`.
        """
        super().__init__(attr=attr, **kwargs)

        # Convert to NXdata
        self.nxclass = "NXdata"
        self.nxname = "data"


class DataProcessed(BaseData):
    """Store processed data in a NeXus NXprocess group."""

    def __init__(
        self,
        program: str = "generic",
        results_name: str = "results",
        serie_name: str = "serie",
        attr: dict = {},
        **kwargs,
    ) -> None:
        """
        Store processed data in a NeXus group.

        The instantiated object will be of type `nexusformat.nexus.NXprocess` [1]. It
        behaves like it, with an additional `create_serie()` method, that adds an NXData
        group.

        It has the following structure :
        ```
        analysis:NXprocess
          date = '2025-08-05T13:28:48.974923'
          program = '{program}'
          {results_name}:NXdata
          {results_name}_{serie_name}{serie_index}:NXdata
          version = '0.1.10'
        ```

        [1] https://nexpy.github.io/nexpy/treeapi.html#nexusformat.nexus.tree.NXprocess

        Parameters
        ----------
        program : str, optional
            Name of the package used to create the object. Must be an installed Python
            package. Default is the `PKG_NAME` global variable.
        results_name : str, optional
            Name of the results Nexus NXData group. Default is 'results'.
        serie_name : str, optional
            Name that will be appended to `results_name`, along with the serie index, to
            build the name of the per-serie Nexus NXData group. Default is 'serie'.
        attr : dict, optional
            Attribute for the NXdata group. Default is an empty dictionary.
        **kwargs : passed to `pyuson.data.DataBase()`.
        """
        self._res_name = results_name
        self._serie_name = serie_name

        super().__init__(attr=attr, **kwargs)

        # Inject method `create_serie()` before NXgroup changes __class__, making any
        # standard instance methods unbound
        nx.NXgroup.create_serie = lambda self, index: _create_serie(self, index)  # ty:ignore[unresolved-attribute]

        # Convert to NXprocess
        self.nxclass = "NXprocess"
        self.nxname = "analysis"

        # Create base datasets
        self["program"] = program
        self["version"] = importlib.metadata.version(program)
        self["results"] = nx.NXdata()

        def _create_serie(self: DataProcessed, index: Any) -> None:
            """
            Create an NXdata group for serie data.

            The name of the created group is : '{results_name}_{serie_name}{index}'.

            Parameters
            ----------
            index : Any
                Something appended to the serie name.
            """
            group_name = f"{self._res_name}_{self._serie_name}{index}"
            if group_name not in self:
                self[group_name] = nx.NXdata()



"""ParameterTree content for pyuson."""

import numpy as np

from pymagnetos.core.gui.widgets import BaseParamContent


class ParamContent(BaseParamContent):
    # List of parameters that will need to be parsed from string to list
    PARAMS_TO_PARSE = ["frame_indices", "range_baseline", "analysis_window"]

    # Define parameters in the "Parameters" section
    children_parameters = [
        dict(
            name="pickup_surface",
            type="float",
            limits=[0, np.inf],
            value=None,
            suffix="m²",
            siPrefix=False,
            readonly=True,
            title="Pickup surface",
        ),
        dict(
            name="pickup_samplerate",
            type="float",
            limits=[0, np.inf],
            value=250e3,
            step=1e3,
            suffix="Hz",
            siPrefix=True,
            readonly=True,
            title="Pickup sample rate",
        ),
        dict(
            name="sample_length",
            type="float",
            limits=[0, np.inf],
            value=None,
            step=0.5e-3,
            suffix="m",
            siPrefix=True,
            readonly=True,
            title="Sample length",
        ),
        dict(
            name="sample_speed",
            type="float",
            limits=[0, np.inf],
            value=None,
            step=0.5e3,
            suffix="m/s",
            siPrefix=True,
            readonly=True,
            title="Sample speed",
        ),
        dict(
            name="detection_mode",
            type="list",
            limits=["reflection", "transmission"],
            value="reflection",
            readonly=True,
            title="Detection mode",
        ),
    ]

    # Define parameters in the "Settings" section (editable)
    children_settings = [
        dict(
            name="echo_index",
            type="int",
            limits=[1, np.inf],
            step=1,
            value=1,
            title="Echo index",
        ),
        dict(name="frame_indices", type="str", title="Displayed frames"),
        dict(
            name="rolling_mean_wlen",
            type="int",
            limits=[0, np.inf],
            step=1,
            value=0,
            title="Rolling mean window length",
        ),
        dict(
            name="rolling_mean_subsample",
            type="bool",
            limits=[True, False],
            value=False,
            title="Rolling mean subsampling",
        ),
        dict(name="range_baseline", type="str", value="", title="Range baseline"),
        dict(
            name="analysis_window", type="str", value="", title="Analysis time window"
        ),
        dict(
            name="max_phase_jump",
            type="float",
            value=0.5,
            step=0.1,
            limits=[0, np.inf],
            title="Max. phase jump (* π)",
        ),
    ]

    children_demodulation = [
        dict(
            name="f0",
            type="float",
            limits=[0, np.inf],
            step=10e6,
            value=0,
            suffix="Hz",
            siPrefix=True,
            title="f0",
        ),
        dict(
            name="auto_find_f0",
            type="bool",
            value=False,
            title="Auto-find f0",
        ),
        dict(
            name="fft_nframes",
            type="int",
            limits=[0, np.inf],
            step=10,
            value=0,
            title="FFT : nframes",
        ),
        dict(
            name="detrend",
            type="bool",
            limits=[False, True],
            value=False,
            title="FFT : detrend",
        ),
        dict(
            name="findsig_nframes",
            type="int",
            limit=[0, np.inf],
            step=10,
            value=0,
            title="Ref. finder : nframes",
        ),
        dict(
            name="findsig_nstd",
            type="float",
            limit=[0, np.inf],
            step=0.1,
            value=1.5,
            title="Ref. finder : nstd",
        ),
        dict(
            name="findsig_extend",
            type="float",
            limit=[-np.inf, np.inf],
            step=0.01,
            value=-0.10,
            title="Ref. finder : ext. frac.",
        ),
        dict(
            name="chunksize",
            type="int",
            limits=[-1, np.inf],
            step=100,
            value=0,
            title="Chunk size",
        ),
        dict(
            name="decimate_factor",
            type="int",
            limits=[0, np.inf],
            step=1,
            value=0,
            title="Decimation factor",
        ),
        dict(
            name="filter_order",
            type="int",
            limits=[1, np.inf],
            step=1,
            value=10,
            title="Filter order",
        ),
        dict(
            name="filter_fc",
            type="float",
            limits=[0, np.inf],
            step=10e6,
            value=300e6,
            suffix="Hz",
            siPrefix=True,
            title="Filter cut-off frequency",
        ),
        dict(name="find_f0", type="action", title="Find f0"),
        dict(name="demodulate", type="action", title="Demodulate"),
    ]



"""Application for TDO experiments during high-magnetic field."""

from PyQt6 import QtWidgets

from . import widgets
from .main import MainWindow

__all__ = ["MainWindow", "run", "widgets"]


def run(exec: bool = True) -> None | MainWindow:
    """
    Build and run the app.

    Set `exec=False` in an interactive prompt to interact with the GUI from the shell.

    Parameters
    ----------
    exec : bool, optional
        Execute the process thread. Set to False in an interactive prompt. Default is
        True.
    """
    app = QtWidgets.QApplication([])
    win = MainWindow()
    win.show()
    if exec:
        app.exec()
    else:
        return win



"""Pydantic models for the EchoProcessor Config."""

from collections.abc import Sequence
from pathlib import Path
from typing import Any

from pydantic import BaseModel, ConfigDict

from ..core import config_models


class Parameters(config_models.Parameters):
    """Parameters section, things related to the experiment itself."""

    pickup_samplerate: float
    sample_length: float
    sample_speed: float
    detection_mode: str
    logamp_slope: float
    rf_offset: float = 0.0


class Settings(BaseModel):
    """Settings section, things related to the analysis."""

    echo_index: int
    frame_indices: Sequence[int]
    rolling_mean_wlen: int
    rolling_mean_subsample: bool
    range_baseline: Sequence[float]
    analysis_window: Sequence[float]
    max_phase_jump: float


class Demodulation(BaseModel):
    """Demodulation section, things related to the digital demodulation process."""

    findsig_nframes: int
    findsig_nstd: float
    findsig_extend: float
    chunksize: int
    decimate_factor: int
    filter_order: int
    filter_fc: float

    f0: float = 0
    auto_find_f0: bool = False
    fft_nframes: int
    detrend: bool


class EchoConfigurationModel(BaseModel):
    """A Model specialized for ultrasound echoes experiment."""

    expid: str
    data_directory: Path

    filenames: dict[str, str | Path] = dict()

    files: dict[str, config_models.File]

    measurements: dict[str, int]

    parameters: Parameters
    settings: Settings

    metadata: config_models.Metadata

    nx: dict[str, dict[str, Any]]
    nexus: config_models.Nexus

    demodulation: Demodulation | None = None

    model_config = ConfigDict(validate_assignment=True)



"""Custom widgets for pytdo, the GUI for TDO experiments."""

from ._buttons import ButtonsWidget
from ._configuration import ConfigurationWidget
from ._graphs import GraphsWidget
from ._param_content import ParamContent

__all__ = ["ButtonsWidget", "ConfigurationWidget", "GraphsWidget", "ParamContent"]



"""The graph area widget for pyuson."""

import pyqtgraph as pg
from PyQt6 import QtCore, QtWidgets
from PyQt6.QtCore import pyqtSignal, pyqtSlot

from pymagnetos.core.gui.widgets import BaseGraphsWidget


class GraphsWidget(BaseGraphsWidget):
    """
    The graphs area with all the plots.

    Signals
    -------
    sig_roi_changed : emits when the draggable time-window moved.
    """

    sig_roi_changed = pyqtSignal()

    def __init__(self) -> None:
        super().__init__()

        # Internal conversion factor from time to frames
        self._time2frame_scale = 1.0

        # Create empty canvases in tabs
        field_tab = self.create_field_plot()
        frame_tab = self.create_frame_plot()
        amplitude_tab = self.create_amplitude_plot()
        phase_tab = self.create_phase_plot()

        self.init_coordinates_on_hover()
        self.disable_autoSI_prefix()

        # Create a main splitter for rows (vertical)
        main_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)

        # Create a top splitter for the first row (horizontal)
        top_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        top_splitter.addWidget(field_tab)
        top_splitter.addWidget(frame_tab)
        top_splitter.setStretchFactor(0, 1)  # field_tab stretch factor
        # frame_tab spans 3 columns
        top_splitter.setStretchFactor(1, 3)

        # Create a bottom splitter for the second row (horizontal)
        bottom_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        bottom_splitter.addWidget(amplitude_tab)
        bottom_splitter.addWidget(phase_tab)
        # amplitude_tab spans 2 columns
        bottom_splitter.setStretchFactor(0, 2)
        # phase_tab spans 2 columns
        bottom_splitter.setStretchFactor(1, 2)

        # Add the top and bottom splitters to the main splitter
        main_splitter.addWidget(top_splitter)
        main_splitter.addWidget(bottom_splitter)
        main_splitter.setStretchFactor(0, 1)  # top row stretch factor
        main_splitter.setStretchFactor(1, 1)  # bottom row stretch factor

        # Set the main splitter as the central widget
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(main_splitter)
        self.setLayout(layout)

    @property
    def time2frame_scale(self) -> float:
        return self._time2frame_scale

    @time2frame_scale.setter
    def time2frame_scale(self, value: float) -> None:
        self._time2frame_scale = value
        self.field.getAxis("top").setScale(self.time2frame_scale)
        self.dfield.getAxis("top").setScale(self.time2frame_scale)

    def create_field_plot(self) -> QtWidgets.QTabWidget:
        tab = QtWidgets.QTabWidget(self)

        # magnetic field (integrated)
        self.field = pg.PlotWidget(title="Magnetic field")
        self.field.setLabel("bottom", "time (s)")
        self.field.setLabel("top", "frame (#) - field not aligned")
        self.field.setLabel("left", "field (T)")
        self.field.showGrid(y=True)
        self._plots_list.add(self.field)

        # pickup coil voltage (measured)
        self.dfield = pg.PlotWidget(title="Pickup")
        self.dfield.setLabel("bottom", "time (s)")
        self.dfield.setLabel("top", "frame (#)")
        self.dfield.setLabel("left", "pickup (V)")
        self.dfield.showGrid(y=True)
        self._plots_list.add(self.dfield)

        tab.addTab(self.field, "B(t)")
        tab.addTab(self.dfield, "Pickup")

        return tab

    def create_frame_plot(self) -> QtWidgets.QTabWidget:
        tab = QtWidgets.QTabWidget(self)

        # amplitude
        self.amp_frame = pg.PlotWidget(title="Frames amplitude")
        self.amp_frame.setLabel("bottom", "time (µs)")
        self.amp_frame.setLabel("left", "signal (V)")
        self.amp_frame.showGrid(y=True)
        self.roi = pg.LinearRegionItem()  # time range selector
        self.roi.sigRegionChangeFinished.connect(self.roi1_changed)
        self.roi.sigRegionChangeFinished.connect(self.sig_roi_changed.emit)
        self.amp_frame.addItem(self.roi, ignoreBounds=True)  # ty:ignore[unknown-argument]
        self._plots_list.add(self.amp_frame)

        # phases
        self.phase_frame = pg.PlotWidget(title="Frames phase")
        self.phase_frame.setLabel("bottom", "time (µs)")
        self.phase_frame.setLabel("left", "signal (V)")
        self.phase_frame.showGrid(y=True)
        self.roi2 = pg.LinearRegionItem()  # twin in phase plot
        self.phase_frame.addItem(self.roi2, ignoreBounds=True)  # ty:ignore[unknown-argument]
        self.roi2.sigRegionChangeFinished.connect(self.roi2_changed)
        self.phase_frame.getPlotItem().addLegend()
        self._plots_list.add(self.phase_frame)

        tab.addTab(self.amp_frame, "Frame amplitude")
        tab.addTab(self.phase_frame, "Frame phase")

        # Keep a reference to it to add more plots if needed
        self.tab_frame_plot = tab

        return tab

    def add_reference_in_frame_tab(self) -> None:
        """Add a plot for the reference signal for digital demodulation."""
        if hasattr(self, "reference_frame") and isinstance(
            self.reference_frame, pg.PlotWidget
        ):
            # Already has a reference frame plot
            return
        self.reference_frame = pg.PlotWidget(title="Frames reference")
        self.reference_frame.setLabel("bottom", "time (µs)")
        self.reference_frame.setLabel("left", "signal (V)")
        self.reference_frame.showGrid(y=True)
        self.disable_autoSI_prefix(self.reference_frame)

        self._plots_list.add(self.reference_frame)
        self.coordinates_on_hover(self.reference_frame)

        self.tab_frame_plot.addTab(self.reference_frame, "Frame reference")

    def create_amplitude_plot(self) -> QtWidgets.QTabWidget:
        tab = QtWidgets.QTabWidget(self)

        # vs field
        self.amp_field = pg.PlotWidget(title="Amplitude (B)")
        self.amp_field.setLabel("bottom", "field (T)")
        self.amp_field.setLabel("left", "attenuation (dB/cm)")
        self.amp_field.showGrid(y=True)
        self.amp_field.getPlotItem().addLegend()
        self._plots_list.add(self.amp_field)

        # vs amplitude
        self.amp_time = pg.PlotWidget(title="Amplitude (t)")
        self.amp_time.setLabel("bottom", "time (s)")
        self.amp_time.setLabel("left", "attenuation (dB/cm)")
        self.amp_time.showGrid(y=True)
        self._plots_list.add(self.amp_time)

        tab.addTab(self.amp_field, "Amplitude (B)")
        tab.addTab(self.amp_time, "Amplitude (t)")

        return tab

    def create_phase_plot(self) -> QtWidgets.QTabWidget:
        tab = QtWidgets.QTabWidget(self)

        # vs field
        self.phase_field = pg.PlotWidget(title="Phase (B)")
        self.phase_field.setLabel("bottom", "field (T)")
        self.phase_field.setLabel("left", "dphi/phi")
        self.phase_field.showGrid(y=True)
        self.phase_field.getPlotItem().addLegend()
        self._plots_list.add(self.phase_field)

        # vs plot
        self.phase_time = pg.PlotWidget(title="Phase (t)")
        self.phase_time.setLabel("bottom", "time (s)")
        self.phase_time.setLabel("left", "dphi/phi")
        self.phase_time.showGrid(y=True)
        self._plots_list.add(self.phase_time)

        tab.addTab(self.phase_field, "Phase (B)")
        tab.addTab(self.phase_time, "Phase (t)")

        return tab

    def init_plot_style(self) -> None:
        """Set up PyQtGraph line styles."""
        w0 = 1
        w1 = 1
        self.pen_field = pg.mkPen("#c7c7c7", width=w0)
        self.pen_amp = pg.mkPen("#1f77b480", width=w0)
        self.pen_amp_demod = pg.mkPen("#d6272880", width=w0)
        self.pen_in_phase = pg.mkPen("#ff7f0e80", width=w1)
        self.pen_out_phase = pg.mkPen("#17becf80", width=w1)
        self.pen_phase_demod = pg.mkPen("#9467bd80", width=w0)
        self.pen_bup = pg.mkPen("#2ca02cbf", width=w0)
        self.pen_bdown = pg.mkPen("#d62728bf", width=w0)

    def init_field_crosshair(self) -> None:
        """Create line cursor on field plot that tracks frame number."""
        self.field_vline = pg.InfiniteLine(angle=90, movable=False)
        self.field.addItem(self.field_vline, ignoreBounds=True)  # ty:ignore[unknown-argument]
        self.field.scene().sigMouseMoved.connect(self.field_crosshair_moved)  # ty:ignore[unresolved-attribute]

    def field_crosshair_moved(self, evt) -> None:
        """Define the callback function when the cursor is moved."""
        pos = evt
        vb = self.field.getViewBox()
        if self.field.sceneBoundingRect().contains(pos):
            mouse_point = vb.mapSceneToView(pos)
            index = int(mouse_point.x() * self.time2frame_scale)
            self.field.setLabel("top", f"frame (#), current : {index}")
            self.field_vline.setPos(mouse_point.x())

    @pyqtSlot()
    def roi1_changed(self) -> None:
        """Update ROI in the phase panel when the ROI in the amplitude panel moved."""
        roi1 = self.roi.getRegion()  # amplitude panel
        roi2 = self.roi2.getRegion()  # phase panel
        if roi1 != roi2:
            self.roi2.setRegion(self.roi.getRegion())

    @pyqtSlot()
    def roi2_changed(self) -> None:
        """Update ROI in the amplitude panel when the ROI in the phase panel moved."""
        roi1 = self.roi.getRegion()  # amplitude panel
        roi2 = self.roi2.getRegion()  # phase panel
        if roi1 != roi2:
            self.roi.setRegion(self.roi2.getRegion())

    def enable_rois(self) -> None:
        """Enable moving ROIs."""
        self.roi.setMovable(True)
        self.roi2.setMovable(True)

    def disable_rois(self) -> None:
        """Disable moving ROIs."""
        self.roi.setMovable(False)
        self.roi2.setMovable(False)



"""
The graphs area that holds all the plots.

The plots themselves should be placed on a custom layout by subclass. This Base class
merely provides plots behavior :
- mouse coordinates on hover,
- CRTL+click to zoom in an area.
"""

from functools import partial

import pyqtgraph as pg
from PyQt6 import QtWidgets
from PyQt6.QtCore import QEvent, Qt
from PyQt6.QtGui import QKeyEvent


class BaseGraphsWidget(QtWidgets.QWidget):
    """Base class for the graphs area with all the plots."""

    _plots_list: set[pg.PlotWidget]

    def __init__(self) -> None:
        super().__init__()

        # Internal flags
        self._mouse_pan_mode = True

        # Registered plots
        self._plots_list = set()

        # Get pens
        self.init_plot_style()

    def init_plot_style(self) -> None:
        """Set and store QPen for styling of curves and ROIs."""
        raise NotImplementedError("Subclass must implement this method.")

    def event(self, a0: QEvent | None = None) -> bool:
        """Handle mouse selection while holding CTRL to make a ROI zoom in plots."""
        # Switch to rectangular zoom selection on the first Ctrl key press event
        if a0 is None:
            return False
        if a0.type() == QKeyEvent.Type.KeyPress and a0.key() == Qt.Key.Key_Control:  # ty:ignore[unresolved-attribute]
            if self._mouse_pan_mode:
                self.switch_mouse_pan_mode(False)

        # Switch back to pan mode on left click when releasing the Ctrl key
        if a0.type() == QKeyEvent.Type.KeyRelease and a0.key() == Qt.Key.Key_Control:  # ty:ignore[unresolved-attribute]
            self.switch_mouse_pan_mode(True)

        return super().event(a0)

    def switch_mouse_pan_mode(self, state: bool) -> None:
        """
        Switch mouse behavior from panning to ROI-zooming when holding CTRL.

        Parameters
        ----------
        state : bool
            If True, all plots are set to pan mode, otherwise they are set in ROI-zoom
            mode.
        """
        # Get mode
        mode = pg.ViewBox.PanMode if state else pg.ViewBox.RectMode
        self._mouse_pan_mode = state

        # Set mode for all registered plots
        for plot in self._plots_list:
            plot.getPlotItem().getViewBox().setMouseMode(mode)

    def disable_autoSI_prefix(self, plot_item: pg.PlotWidget | None = None) -> None:
        """Disable replacing units in power of 10."""
        if plot_item is None:
            for plot in self._plots_list:
                plot.getAxis("left").enableAutoSIPrefix(False)
                plot.getAxis("bottom").enableAutoSIPrefix(False)
        else:
            plot_item.getAxis("left").enableAutoSIPrefix(False)
            plot_item.getAxis("left").enableAutoSIPrefix(False)

    def init_coordinates_on_hover(self) -> None:
        """Add x, y coordinates on mouse hover for all registered plots."""
        for plot in self._plots_list:
            self.coordinates_on_hover(plot)

    def coordinates_on_hover(self, plot) -> None:
        """Add x, y coordinates on mouse hover."""
        plot.scene().sigMouseMoved.connect(partial(self.mouse_moved_in_plot, plot))
        plot.setLabel("top", "x=, y=")
        plot.getAxis("top").setStyle(showValues=False)

    def mouse_moved_in_plot(self, plot, evt) -> None:
        """Define what happens when the mouse moves within a plot."""
        pos = evt
        if plot.getPlotItem().sceneBoundingRect().contains(pos):
            vb = plot.getPlotItem().getViewBox()
            mouse_point = vb.mapSceneToView(pos)
            label_text = f"x={mouse_point.x():0.6f}, y={mouse_point.y():0.6f}"
            plot.setLabel("top", label_text)

    def clear_all_plots(self) -> None:
        """Clear plots."""
        for plot in self._plots_list:
            plot.clearPlots()



"""
Application for ultrasound echoes analysis.

This is the main window definition, built with the custom widgets found in the widgets
folder.
"""

import importlib.metadata
from collections.abc import Sequence
from pathlib import Path

import pyqtgraph as pg
from PyQt6 import QtGui, QtWidgets
from PyQt6.QtCore import pyqtSignal as Signal
from PyQt6.QtCore import pyqtSlot as Slot

from pymagnetos import sp
from pymagnetos.core.gui import BaseMainWindow, widgets

from .widgets import (
    BatchProcessingWidget,
    ButtonsWidget,
    ConfigurationWidget,
    GraphsWidget,
    ParamContent,
)
from .worker import Worker

ICON_PATH = str(Path(__file__).parent / "assets" / "icon.png")
REGEXP_EXPID_SEPARATORS = r"[_-]"
PROGRAM_NAME = "pymagnetos"
ALLOWED_FORMAT = ("toml", "json", "nx5", "nxs", "h5", "hdf5")
LOG_LEVEL = "INFO"
VERSION = importlib.metadata.version(PROGRAM_NAME)

PARAM_ANALYSIS_TRIGGER = ["range_baseline", "max_phase_jump"]


class MainWindow(BaseMainWindow):
    """
    A graphical user interface application for ultrasound echoes analysis.

    The `MainWindow` class defines the main thread with the front-end user interface.

    It is built with the custom components defined in the `widgets` module. Those
    components are referenced in the main thread with a leading `w` :
    `self.wconfiguration` : the "Configuration" tab
    `self.wfiles` : the "Files" tab
    `self.wbatch` : the "Batch processing" tab
    `self.wgraphs` : the widget holding all the graphs
    `self.wbuttons` : the widget holding the main buttons
    `self.wlog` : the widget that streams the log output

    Each of those components exposes signals, to which the main thread connects its own
    methods, instead of connecting directly to the widgets (buttons, checkboxes and the
    like). Those signals are named with a trailing `sig_` so that makes it easy to know
    exactly what should be connected in the main thread.

    When loading a configuration (or a NeXus) file, the main threads creates a worker
    thread (`Worker`). The latter instantiates an `EchoProcessor` object that
    perform the actual analysis.

    Additional signals
    ------------------
    All signals emit in their corresponding methods and are connected to the
    corresponding methods in the worker thread.

    `sig_worker_rolling`
    `sig_worker_find_f0`
    `sig_worker_export_csv`
    """

    sig_worker_rolling = Signal()
    sig_worker_find_f0 = Signal()
    sig_worker_export_csv = Signal(str, bool)

    wbatch: BatchProcessingWidget
    wbuttons: ButtonsWidget
    wconfiguration: ConfigurationWidget
    wgraphs: GraphsWidget

    worker: Worker

    _window_icon = ICON_PATH

    def __init__(self):
        # Register the widgets to use
        self._type_wbatch = BatchProcessingWidget
        self._type_wbuttons = ButtonsWidget
        self._type_wconfiguration = ConfigurationWidget
        self._type_wfiles = widgets.FileBrowserWidget
        self._type_wgraphs = GraphsWidget
        self._param_content = ParamContent
        self._type_worker = Worker

        super().__init__()

        # Additional flags
        self._is_demod_connected = False  # demodulation-related signals are connected

        # Initialize window
        self.setGeometry(300, 300, 900, 450)
        self.setWindowTitle("EchoAnalyzer")
        self.setWindowIcon(QtGui.QIcon(self._window_icon))

        self.logger.info(f"Running pymagnetos.pyuson v{VERSION}")

    def init_demodulation(self):
        """
        Additionnal initializations for digital demodulation mode.

        Add the Demodulation section in the parameter tree and the frame reference plot.

        It happens only when a worker is created so we connect its signals as well,
        instead of in `connect_worker()` for the other buttons.
        """
        # Add the Demodulation section to ParameterTree
        self.wconfiguration.init_demodulation_tree()

        # Update the files filter suffix in the batch processing panel
        self.wbatch.line_suffix.setText(".wfm")
        self.wbatch.add_findf0_checkbox()

        # Fill the demodulation parameters from the worker
        for p in self.wconfiguration.demodulation_parameters:
            self.set_param_from_worker(p.name(), context="demodulation")

        # Add the frame reference signal plot tab
        self.wgraphs.add_reference_in_frame_tab()

        # Connect the worker signals
        self.sig_worker_find_f0.connect(self.worker.find_f0)
        self.worker.sig_find_f0_finished.connect(self.find_f0_finished)

        # Connect other signals if there were not already
        if not self._is_demod_connected:
            self.wconfiguration.sig_find_f0.connect(self.find_f0)
            self.wconfiguration.sig_demodulate.connect(self.preprocess)
            self._is_demod_connected = True

    def init_batch_processing(self):
        """
        Create the "Batch processing" tab.

        It's multi-panel file picker, where the user can queue files to be
        batch-processed with the same settings.
        """
        super().init_batch_processing()

        # Connect
        self.wbatch.sig_echo_index_changed.connect(self.update_echo_index_from_batch)

    def init_buttons(self):
        """
        Create and connect the main action buttons.

        The created buttons and their functions are :
        Load Data : load raw data
        Show Frames : display raw data from frames specified in the settings
        Rolling average : apply rolling average
        Refresh : re-compute the average, attenuation and phase-shift
        Export as CSV : export the results for the current echo index as a CSV file
        Save as NeXus : export all data as a single NeXus (hdf5) file
        """
        super().init_buttons()

        # Connect
        self.wbuttons.sig_show.connect(self.show_frames)
        self.wbuttons.sig_rolling.connect(self.rolling_average)
        self.wbuttons.sig_refresh.connect(self.handle_roi_changed)
        self.wbuttons.sig_save_csv.connect(self.export_csv)

    def init_plots(self):
        """
        Create the graph area that holds all the plots.

        The user-draggable ROI defining the time window in which the data is averaged is
        connected here.
        """
        super().init_plots()

        # Connect
        self.wgraphs.sig_roi_changed.connect(self.handle_roi_changed)

    def init_log(self):
        """
        Create the widget where the log will be printed.

        The logger is shared with the `pyuson` package-level logger so stream is also
        printed in stdout and the log file.
        """
        super()._init_log(PROGRAM_NAME, log_level=LOG_LEVEL)

    @Slot()
    def load_file(self):
        """
        Load a configuration or NeXus file and start a worker.

        Callback for when the "File" parameter changes, or a file is drag & dropped in
        the window.
        """
        super().load_file()
        if self.worker.proc.is_digital:
            self.init_demodulation()

        # Set the ROI
        if self.worker.proc.is_nexus_file:
            # Turn off analysis in case it is already in the NeXus file
            self._do_analysis = False
        region = self.worker.proc.analysis_window
        if isinstance(region, Sequence) and len(region) == 2:
            self.update_roi(region)

        if self.worker.is_data_loaded:
            # If data was loaded automatically, we can show it
            self.load_data_finished()

        if self.worker.proc.is_nexus_file:
            # If NeXus file, data and results can be shown
            self.load_data_finished()
            self.run_analysis_finished()
            self._do_analysis = True  # turn analysis back on

    def set_param_from_worker(self, param_name: str, context: str):
        """
        Set a parameter read from the `Config` object of the `EchoProcessor` worker.

        This override adds the "demodulation" section of the configuration.

        Parameters
        ----------
        param_name : str
            Name of the parameter in the tree and `Config`.
        context : {"parameters", "settings", "demodulation"}
            Configuration file section.
        """
        if context == "demodulation":
            if hasattr(self.worker.proc.cfg.demodulation, param_name):
                config_value = getattr(self.worker.proc.cfg.demodulation, param_name)
                if param_name in self.wconfiguration.parameters_to_parse:
                    # special case that needs to be converted to string
                    config_value = self.wconfiguration.get_numbers_from_text(
                        config_value
                    )

                self.wconfiguration.demodulation_parameters[param_name] = config_value
        else:
            super().set_param_from_worker(param_name, context)

    def connect_worker(self):
        """
        Connect additional signals from the main thread to tasks in the worker.

        Note that the demodulation tasks are initialized in the
        `init_parameter_tree_demodulation()` method instead.
        """
        super().connect_worker()

        # Rolling average
        self.sig_worker_rolling.connect(self.worker.rolling_average)
        self.worker.sig_rolling_finished.connect(self.rolling_average_finished)

        # Export as CSV
        self.sig_worker_export_csv.connect(self.worker.export_as_csv)
        self.worker.sig_export_csv_finished.connect(self.export_csv_finished)

    @Slot(str, str, object)
    def set_worker_config_from_tree(
        self, context: str, param_name: str, value: float | int | list | bool | str
    ):
        """
        Read parameter from the Parameter Tree and set its sibling in the Config object.

        This override adds support for the "demodulation" section of the configuration.

        Parameters
        ----------
        param_name : str
            Name of the parameter in the Parameter Tree.
        context : {"parameters", "settings", "demodulation"}
            Define where the key is stored in the Config object (corresponding to the
            configuration file section).
        """
        if context == "demodulation":
            setattr(self.worker.proc.cfg.demodulation, param_name, value)
        else:
            super().set_worker_config_from_tree(context, param_name, value)

    def run_after_tree_change(
        self, context: str, param_name: str, value: float | int | list | bool | str
    ) -> None:
        """Run additional actions after a parameter change in the parameter tree."""
        if context != "settings":
            return

        match param_name:
            case "analysis_window":
                if isinstance(value, list):
                    self.update_roi(value)
            case param_name if param_name in PARAM_ANALYSIS_TRIGGER:
                self.run_analysis()
            case "echo_index":
                self.update_echo_index_from_settings()
                self.run_analysis()
            case "frame_indices":
                self.show_frames()

    @Slot()
    def load_data_finished(self):
        """
        Show frames after data is loaded.

        Callback for when the worker has finished loading data.
        """
        self.show_frames()
        self.align_field_finished()
        self.handle_roi_changed()
        self.wconfiguration.host_parameters["rf_frequency"] = self.worker.proc.metadata[
            "rf_frequency"
        ]
        super().load_data_finished()

    @Slot()
    def align_field_finished(self):
        """
        Re-plot magnetic field with its new time vector.

        Set the crosshair that tracks frames in the plot on mouse hover, and store
        increasing and decreasing magnetic field indices to plot them in different
        colors.

        Callback for when the worker has finished loading data. This is re-triggered
        everytime the time vector is susceptible to change (subsampling, ...).
        """
        # Update field plot axis label
        self.wgraphs.field.setLabel("top", "frame (#)")

        super().align_field_finished()

    @Slot()
    def find_f0(self):
        """
        Find center frequency.

        Send the signal to the worker to find the radio-frequency from the signal (or
        the configuration file). Callback for the "Find f0" button in the demodulation
        section of the parameter tree.
        """
        if not self.check_data_loaded():
            return
        self.disable_buttons()
        self.sig_worker_find_f0.emit()

    @Slot(float)
    def find_f0_finished(self, f0: float):
        """
        Update f0 in the parameter tree.

        Callback for when the worker has finished the frequency detection. `f0` is
        passed with the signal.
        """
        self.wconfiguration.demodulation_parameters["f0"] = f0

        self.enable_buttons()

    @Slot()
    def preprocess(self):
        """
        Run the digital demodulation.

        Create a progress bar and send the signal to the worker to perform the digital
        demodulation. Callback for the "Demodulate" button in the "Demodulation" section
        of the parameter tree.
        """
        if not self.check_data_loaded():
            return

        # Create the progress bar
        nchunks = self.worker.proc.nframes
        self.pbar_demodulation = widgets.PopupProgressBar(
            min=0, max=nchunks, title="Digital demodulation"
        )
        self.worker.sig_preprocess_progress.connect(
            self.pbar_demodulation.update_progress
        )
        self.pbar_demodulation.start_progress()

        super().preprocess()

    @Slot()
    def preprocess_finished(self):
        """
        Show the frames demodulated signal and run computations after demodulation.

        Callback for when the worker has finished the digital demodulation.
        """
        self.wconfiguration.demodulation_parameters["f0"] = (
            self.worker.proc.cfg.demodulation.f0
        )
        self.pbar_demodulation.close()
        self.worker.sig_preprocess_progress.disconnect()

        # Rolling average is resetted and signal might be subsampled
        self.align_field_finished()
        self.show_frames()
        self.handle_roi_changed()

        super().preprocess_finished()

    @Slot()
    def run_analysis_finished(self):
        """
        Update the plots with the new attenuation and phase-shift.

        Callback for when the worker has finished averaging and computing. `status` is
        passed with the pyqtSignal, it tells if the averaging was successful.
        """
        if not self.worker.proc.is_averaged:
            self.enable_buttons()
            self.logger.error("[GUI] Averaging frames failed, check the messages above")
            return

        # Check the magnetic field is aligned
        if not self.check_field_aligned():
            # Force align
            self.worker.align_field()

        # Plot results
        # Attenuation
        # Versus magnetic field (dB/cm)
        self.plot_var_field(self.wgraphs.amp_field, "attenuation", mult=1e-2)
        # Versus time (dB/cm)
        self.plot_var_time(self.wgraphs.amp_time, "attenuation", mult=1e-2)

        # Phase-shift
        # Versus magnetic field
        self.plot_var_field(self.wgraphs.phase_field, "phaseshift")
        # Versus time
        self.plot_var_time(self.wgraphs.phase_time, "phaseshift")

        super().run_analysis_finished()

    @Slot()
    def rolling_average(self):
        """
        Perform rolling average.

        Send the signal to the worker to perform the rolling average. Callback for the
        "Rolling average" button.
        """
        if not self.check_data_loaded():
            return
        self.disable_buttons()
        self.sig_worker_rolling.emit()

    @Slot()
    def rolling_average_finished(self):
        """
        Plot magnetic field in case it was subsampled and retrigger computations.

        Callback for the when the worker has finished the rolling average.
        """
        self.align_field_finished()
        self.handle_roi_changed()
        self.enable_buttons()

    @Slot()
    def export_csv(self):
        """
        Export results for the current echo index as CSV.

        Generate a default file name and open a file picker dialog for the user to
        choose an output file. Determine if the attenuation should be converted to
        dB/cm, and send the signal to the worker to save the results as CSV.
        Callback for the "Export as CSV" button.
        """
        if not self.check_data_loaded():
            return

        self.disable_buttons()

        default_fname = self.worker.proc.get_csv_filename()
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Save current echo index as...",
            default_fname,
            "Text files (*.txt, *.csv, *.tsv)",
        )

        if fname:
            self.sig_worker_export_csv.emit(
                fname, self.wbuttons.checkbox_to_cm.isChecked()
            )
        else:
            self.logger.error("[GUI] Invalid output file name for CSV file")
            self.export_csv_finished()

    @Slot()
    def export_csv_finished(self):
        """Callback for when the worker has finished exporting as CSV."""
        self.enable_buttons()

    @Slot()
    def batch_process(self):
        """
        Run batch processing on selected files.

        List files in the queue, determine corresponding experiment IDs, create a
        progress bar and send the signal to the worker to perform the batch-processing.
        Callback for the "Batch process" button in the "Batch processing" tab.
        """
        super().batch_process()

        # Get settings
        rolling_average = self.wbatch.checkbox_rolling_average.isChecked()
        to_csv = self.wbatch.checkbox_save_as_csv.isChecked()
        to_cm = self.wbuttons.checkbox_to_cm.isChecked()
        if hasattr(self.wbatch, "checkbox_find_f0"):
            find_f0 = self.wbatch.checkbox_find_f0.isChecked()
        else:
            find_f0 = False

        # Set up progress bar
        self.pbar_batch = widgets.PopupProgressBar(
            min=0, max=len(self.batch_expids), title="Batch processing"
        )
        self.worker.sig_batch_progress.connect(self.batch_process_step_finished)
        self.pbar_batch.start_progress()

        # Run the batch processing
        self.disable_buttons()
        self.sig_worker_batch.emit(
            (self.batch_expids, rolling_average, to_csv, to_cm, find_f0)
        )

    @Slot(int)
    def batch_process_step_finished(self, idx: int):
        """Track progress of the batch processing."""
        self.pbar_batch.update_progress(idx)
        super().batch_process_step_finished(idx)

    @Slot()
    def batch_process_finished(self):
        """
        Cleanup after batch-processing.

        Callback for when the worker has finished the batch-processing.
        """
        self.pbar_batch.close()
        super().batch_process_finished()

    @Slot()
    def handle_roi_changed(self):
        """
        Define what to do when the ROI in the graph is changed.

        The "analysis_window" setting is updated in the tree, a new analysis is requested.
        """
        # Get time range
        xmin, xmax = self.wgraphs.roi.getRegion()

        # Check the ROI was changed since it was created
        if (xmin, xmax) == (0, 1):
            return

        # Update parameter in the worker directly to ensure real values are used without
        # parsing. This is because in the tree the range is displayed as a string.
        if self.check_config_loaded():
            self.worker.proc.analysis_window = [xmin, xmax]

        # The displayed numbers have 16 digits, so trim it, without updating the worker
        self._ignore_tree_change = True
        self.wconfiguration.settings_parameters["analysis_window"] = (
            self.wconfiguration.get_text_from_numbers([xmin, xmax], precision=6)
        )
        self._ignore_tree_change = False

        # Average frames and compute attenuation and phase-shift
        self.run_analysis()

    def update_roi(self, new_region: Sequence[float | int]):
        """Update the ROI in the graph from "Analysis time window" in the parameter tree."""
        if len(new_region) != 2:
            self.logger.error(f"[GUI] Invalid analysis window : {new_region}")
            return
        elif new_region[0] >= new_region[1]:
            self.logger.error(f"[GUI] Invalid analysis window : {new_region}")
            return

        # Update ROI, if it changes, it will trigger the analysis through a call to to
        # handle_roi_changed()
        self.wgraphs.roi.setRegion(new_region)

    def plot_field(self):
        """
        Display magnetic field versus time.

        The pickup coil voltage and the magnetic field are shown. If it is aligned with
        the experiment time vector, a cross-hair tracks the frame number on mouse hover.
        """
        super().plot_field()

        # If data was loaded, we can show frames indices on top with a crosshair
        if self.check_data_loaded():
            self.wgraphs.time2frame_scale = (
                self.worker.proc.get_data_processed("time_exp").size
                / self.worker.proc.get_data_processed("magfield_time")[-1]
            )

            # Create the crosshair if it does not exist yet
            if not hasattr(self.wgraphs, "field_vline"):
                self.wgraphs.init_field_crosshair()

    @Slot()
    def show_frames(self):
        """
        Show some frames echoes, in amplitude and phase.

        Only the data for the frames specified in the settings are shown. It might seem
        convoluted because it is needed to fetch the correct dataset names, depending on
        whether we're in analog or digital mode.

        Callback for the "Show frames" button, also triggered whenever new data is
        loaded (raw or demodulated).
        """
        if not self.check_data_loaded():
            return
        elif "time_meas" not in self.worker.proc.data_raw:
            self.logger.warning(
                "[GUI] Data was not properly loaded, check messages above"
            )
            return

        # Clear existing traces
        self.wgraphs.amp_frame.clearPlots()
        self.wgraphs.phase_frame.clearPlots()
        if self.worker.proc.is_digital:
            self.wgraphs.reference_frame.clearPlots()

        frame_indices = self.worker.proc.cfg.settings.frame_indices
        for ind in frame_indices:
            idx = int(ind)

            # Set the curve legend only for the last one
            if idx == frame_indices[-1]:
                legend_flag = True
            else:
                legend_flag = False

            if self.worker.proc.is_digital:
                # Plot raw trace
                self.wgraphs.amp_frame.plot(
                    self.worker.proc.get_data_raw("time_meas"),
                    self.worker.proc.get_data_raw(self.worker.proc._sig_name)[:, idx],
                    pen=self.wgraphs.pen_amp,
                )
                # Plot demodulated results if it exists
                iname = self.worker.proc.measurements[0]
                qname = self.worker.proc.measurements[1]
                if self.worker.proc.get_data_processed(
                    iname, checkonly=True
                ) and self.worker.proc.get_data_processed(qname, checkonly=True):
                    self.show_frames_demodulated()
                # Plot reference trace
                self.wgraphs.reference_frame.plot(
                    self.worker.proc.get_data_raw("time_meas"),
                    self.worker.proc.get_data_raw("reference")[:, idx],
                    pen=pg.mkPen(color=pg.intColor(idx), hues=len(frame_indices)),
                )
            else:
                # Analog mode
                aname = self.worker.proc.measurements[0]
                self.wgraphs.amp_frame.plot(
                    self.worker.proc.get_data_raw("time_meas"),
                    self.worker.proc.get_data_raw(aname)[:, idx],
                    pen=self.wgraphs.pen_amp,
                )
                iname = self.worker.proc.measurements[1]
                self.wgraphs.phase_frame.plot(
                    self.worker.proc.get_data_raw("time_meas"),
                    self.worker.proc.get_data_raw(iname)[:, idx],
                    pen=self.wgraphs.pen_in_phase,
                    name="in-phase" if legend_flag else None,
                )
                qname = self.worker.proc.measurements[2]
                self.wgraphs.phase_frame.plot(
                    self.worker.proc.get_data_raw("time_meas"),
                    self.worker.proc.get_data_raw(qname)[:, idx],
                    pen=self.wgraphs.pen_out_phase,
                    name="out-of-phase" if legend_flag else None,
                )

    def show_frames_demodulated(self):
        """Overlay demodulation result on raw frames."""
        # Get measurements names
        iname = self.worker.proc.measurements[0]
        qname = self.worker.proc.measurements[1]

        # Get correct time vector if decimation was used
        if self.worker.proc.is_decimated:
            xt = self.worker.proc.get_data_processed("time_meas")
        else:
            xt = self.worker.proc.get_data_raw("time_meas")

        # Change labels
        self.wgraphs.phase_frame.setLabel("left", "phase (rad)")

        # Compute amplitude and phase of demodulated signal
        frames_ind = [int(ind) for ind in self.worker.proc.cfg.settings.frame_indices]
        amplitude = sp.compute_amp_iq(
            self.worker.proc.get_data_processed(iname)[:, frames_ind],
            self.worker.proc.get_data_processed(qname)[:, frames_ind],
        )
        phase = sp.compute_phase_iq(
            self.worker.proc.get_data_processed(iname)[:, frames_ind],
            self.worker.proc.get_data_processed(qname)[:, frames_ind],
        )

        # Rescale amplitude to show it with the raw trace on the same scale
        amplitude = sp.rescale_a2b(
            amplitude,
            self.worker.proc.get_data_raw(self.worker.proc._sig_name)[:, frames_ind],
        )

        for ind in range(len(frames_ind)):
            idx = int(ind)
            self.wgraphs.amp_frame.plot(
                xt,
                amplitude[:, idx],
                pen=self.wgraphs.pen_amp_demod,
            )
            self.wgraphs.phase_frame.plot(
                xt,
                phase[:, idx],
                pen=self.wgraphs.pen_phase_demod,
            )

    @Slot()
    def update_echo_index_from_batch(self):
        """Update echo index in the parameter tree when changed from the Batch tab."""
        self.wconfiguration.settings_parameters["echo_index"] = self.wbatch.echo_index

    @Slot()
    def update_echo_index_from_settings(self):
        """Update echo index in the Batch tab when changed from the parameter tree."""
        self.wbatch.echo_index = self.wconfiguration.settings_parameters["echo_index"]

    def reset(self):
        """Quit and delete worker and thread, resetting plots and parameters."""
        # Quit thread
        super().reset()

        # Reset demodulation parameter section
        if (
            hasattr(self.wconfiguration, "demodulation_parameters")
            and self.wconfiguration.demodulation_parameters is not None
        ):
            self.wconfiguration.host_parameters.removeChild(
                self.wconfiguration.demodulation_parameters
            )
            self.wconfiguration.demodulation_parameters = None
        # Remove the "Find f0" checkbox in the batch processing tab
        if hasattr(self.wbatch, "checkbox_find_f0"):
            self.wbatch.buttons_grid.removeWidget(self.wbatch.checkbox_find_f0)
            self.wbatch.checkbox_find_f0.deleteLater()
            self.wbatch.checkbox_find_f0 = None



"""
Shared Pydantic models for configuration.

Each model corresponds to a section in the configuration file.
"""

from typing import Any, Literal

from pydantic import BaseModel


class File(BaseModel):
    """Binary or text file."""

    ext: str
    header: int

    # For binary file
    precision: int | None = None
    endian: Literal["<", ">"] | None = None
    order: Literal["C", "F"] | None = None

    # For text file
    delimiter: str | None = None


class Parameters(BaseModel):
    """Parameters section, things related to the experiment itself."""

    pickup_surface: float
    pickup_number: int
    pickup_index: int


class Metadata(BaseModel):
    """Metadata section, defining how to read the metadata from files."""

    index_map: dict[str, int]
    conversion_map: dict[str, bool]


class Nexus(BaseModel):
    """The NeXus section, that sets the HDF5 group and dataset base names."""

    groups: dict[str, dict[str, Any]]
    datasets: dict[str, dict[str, Any]]



"""Pydantic models for the LinxProcessor Config."""

from pathlib import Path
from typing import Any, Literal

from pydantic import BaseModel, ConfigDict

from ..core import config_models


class Parameters(config_models.Parameters):
    """Parameters section, things related to the experiment itself."""

    samplerate: float
    current: float
    gain: float
    gain_rms: bool


class Settings(BaseModel):
    """Settings section, things related to the analysis."""

    ref_name: str
    signal_name: str

    detrend_loc: Literal["end", "start", "start+end", "indices"]
    detrend_prct: float
    time_offset: float

    phase_fit_ncycles: int
    filter_order: int
    filter_fc: float
    filter_edge: Literal["pad", "gust"]
    filter_padding_type: Literal["even", "odd", "constant"]
    filter_padding_length: int
    filter_ir_length: int
    
    filter_passband_mult: float
    filter_stopband_mult: float
    filter_passband_att: float
    filter_stopband_att_mult: float
    
    f0: float
    phase_offset: float
    auto_phase_method: Literal["at peak field", "minimum oop"]


class LinxConfigurationModel(BaseModel):
    """A Model specialized for digital lock-in."""

    expid: str
    data_directory: Path

    filenames: dict[str, str | Path] = dict()

    files: dict[str, config_models.File]

    measurements: dict[str, int]

    parameters: Parameters
    settings: Settings

    nx: dict[str, dict[str, Any]]
    nexus: config_models.Nexus

    model_config = ConfigDict(validate_assignment=True)



"""
Define various parameters to set in a PyQtGraph Parameter Tree.

To be able to automatically update them from the configuration file, their name should
match exactly the one in the configuration file.
Use the "title" kwarg to set the displayed name on the widget.

It is used in the ConfigurationWidget.
"""


class BaseParamContent:
    PARAMS_TO_PARSE: list[str]
    children_files: list[dict[str, str | bool]] = [
        dict(name="file", type="file", value="", title="Configuration file"),
        dict(name="expid", type="str", value="", readonly=False, title="Experiment ID"),
        dict(name="autoload", type="bool", value=True, title="Load data automatically"),
    ]
    children_parameters: list[dict[str, str | bool | list | float | int | None]]
    children_settings: list[dict[str, str | bool | list | float | int | None]]



"""The graphs area that holds all the plots."""

import pyqtgraph as pg
from PyQt6 import QtCore, QtWidgets
from PyQt6.QtCore import pyqtSignal, pyqtSlot

from pymagnetos.core.gui.widgets import BaseGraphsWidget


class GraphsWidget(BaseGraphsWidget):
    """
    The graphs area with all the plots.

    There are two tabs, one with :
    - dB/dt
    - B(t)
    - TDO(t)
    - TDO(B)
    The other with :
    - B(t)
    - TDO(B)
    - TDO_detrended(B)
    - TDO_detrended(1/B)
    - FFT

    Since B(t) and TDO(B) are re-used, placeholders plots are created. They are replaced
    when the tab is selected.

    Signals
    -------
    sig_roi1_changed : emits when the draggable field-window moved in tdo(B).
    sig_roi2_changed : emits when the draggable field-window moved in tdo_detrend(B).
    """

    sig_roi1_changed = pyqtSignal()
    sig_roi2_changed = pyqtSignal()

    def __init__(self):
        super().__init__()
        # Flags
        self._sync_roi = False

        # Create layouts
        widget1 = self.create_first_tab()
        widget2 = self.create_second_tab()

        # Create the tabs
        tab = QtWidgets.QTabWidget(self)
        tab.addTab(widget1, "TDO")
        tab.addTab(widget2, "Oscillations")
        tab.currentChanged.connect(self.move_plots)  # replace placeholders

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(tab)
        self.setLayout(layout)

        # Apply axes style
        self.customize_axis()
        self.init_coordinates_on_hover()
        self.disable_autoSI_prefix()

    def create_first_tab(self) -> QtWidgets.QSplitter:
        """Create the tab with the magnetic field and TDO signal."""
        # Create the plots
        self.create_field_plot()
        self.create_signal_plot()

        # Create splitter for the top row
        self.tab1_top_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        self.tab1_top_splitter.addWidget(self.dfield)
        self.tab1_top_splitter.addWidget(self.field)
        self.tab1_top_splitter.addWidget(self.sig_time)

        # Create splitter for the bottom row
        self.tab1_main_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        self.tab1_main_splitter.addWidget(self.tab1_top_splitter)
        self.tab1_main_splitter.addWidget(self.sig_field)

        # Adjust ratios
        self.tab1_main_splitter.setStretchFactor(0, 1)
        self.tab1_main_splitter.setStretchFactor(1, 2)

        return self.tab1_main_splitter

    def create_second_tab(self) -> QtWidgets.QSplitter:
        """Create the tab with the detrended TDO signals and the FFT."""
        self.create_tdo_plot()
        self.create_fft_plot()

        # Create splitter for the top row
        self.tab2_top_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        # Add the placeholders plots, they will be replaced when the tab is shown
        self.tab2_top_splitter.addWidget(self._v2field)
        self.tab2_top_splitter.addWidget(self._v2sig_field)

        # Create splitter for the bottom row
        bottom_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        bottom_splitter.addWidget(self.tdo_field)
        bottom_splitter.addWidget(self.tdo_inverse_field)
        bottom_splitter.addWidget(self.fft)

        # Create main splitter
        self.tab2_main_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        self.tab2_main_splitter.addWidget(self.tab2_top_splitter)
        self.tab2_main_splitter.addWidget(bottom_splitter)

        return self.tab2_main_splitter

    def create_field_plot(self) -> None:
        # magnetic field (integrated)
        self.field = pg.PlotWidget(title="Magnetic field")
        self.field.setLabel("bottom", "time (s)")
        self.field.setLabel("left", "field (T)")
        
        self._plots_list.add(self.field)
        # magnetic field (placeholders)
        self._v1field = pg.PlotWidget(title="Magnetic field")
        self._v1field.setLabel("bottom", "time (s)")
        self._v1field.setLabel("left", "field (T)")
        self._v2field = pg.PlotWidget(title="Magnetic field")
        self._v2field.setLabel("bottom", "time (s)")
        self._v2field.setLabel("left", "field (T)")

        # pickup coil voltage (measured)
        self.dfield = pg.PlotWidget(title="Pickup")
        self.dfield.setLabel("bottom", "time (s)")
        self.dfield.setLabel("left", "pickup (V)")
        self._plots_list.add(self.dfield)

    def create_signal_plot(self):
        # TDO signal versus field
        self.sig_field = pg.PlotWidget(title="Signal versus field")
        self.sig_field.setLabel("bottom", "field (T)")
        self.sig_field.setLabel("left", "TDO frequency (Hz)")
        # Field-window selector
        self.roi = pg.LinearRegionItem(
            brush=pg.mkBrush("#0000ff1a"), pen=self.pen_fitbounds
        )
        # Disable mouse drag
        self.roi.hoverEvent = lambda *args, **kwargs: None  # ty:ignore[invalid-assignment]
        self.roi.mouseDragEvent = lambda *args, **kwargs: None  # ty:ignore[invalid-assignment]
        self.sig_field.addItem(self.roi, ignoreBounds=True)  # ty:ignore[unknown-argument]

        self.roi.sigRegionChangeFinished.connect(self.roi1_changed)
        self.sig_field.getPlotItem().addLegend()
        self._plots_list.add(self.sig_field)

        # TDO signal versus field (placeholders)
        self._v1sig_field = pg.PlotWidget(title="Signal versus field")
        self._v1sig_field.setLabel("bottom", "field (T)")
        self._v1sig_field.setLabel("left", "TDO frequency (Hz)")
        self._v2sig_field = pg.PlotWidget(title="Signal versus field")
        self._v2sig_field.setLabel("bottom", "field (T)")
        self._v2sig_field.setLabel("left", "TDO frequency (Hz)")

        # TDO signal versus time
        self.sig_time = pg.PlotWidget(title="Signal versus time")
        self.sig_time.setLabel("bottom", "time (s)")
        self.sig_time.setLabel("left", "TDO frequency (Hz)")
        self._plots_list.add(self.sig_time)

    def create_tdo_plot(self):
        # vs field
        self.tdo_field = pg.PlotWidget(title="Oscillatory part versus field")
        self.tdo_field.setLabel("bottom", "field (T)")
        self.tdo_field.setLabel("left", "TDO detrended")
        self.tdo_field.getPlotItem().addLegend()
        # Field-window selector for FFT range
        self.roi2 = pg.LinearRegionItem(
            brush=pg.mkBrush("#0000ff1a"), pen=self.pen_fftbounds
        )
        # Disable mouse drag
        self.roi2.hoverEvent = lambda *args, **kwargs: None  # ty:ignore[invalid-assignment]
        self.roi2.mouseDragEvent = lambda *args, **kwargs: None  # ty:ignore[invalid-assignment]

        self.tdo_field.addItem(self.roi2, ignoreBounds=True)  # ty:ignore[unknown-argument]
        self.roi2.sigRegionChangeFinished.connect(self.roi2_changed)

        # Vertical lines to show polynomial fit range
        self.fit_bounds1 = pg.InfiniteLine(pen=self.pen_fitbounds, movable=False)
        self.fit_bounds2 = pg.InfiniteLine(pen=self.pen_fitbounds, movable=False)
        self.tdo_field.addItem(self.fit_bounds1)
        self.tdo_field.addItem(self.fit_bounds2)
        self._plots_list.add(self.tdo_field)

        # vs 1/B
        self.tdo_inverse_field = pg.PlotWidget(title="Oscillatory part versus 1/B")
        self.tdo_inverse_field.setLabel("bottom", "1/B (T^-1)")
        self.tdo_inverse_field.setLabel("left", "TDO detrended")
        # Vertical lines to show FFT range
        self.fft_bounds1 = pg.InfiniteLine(pen=self.pen_fftbounds, movable=False)
        self.fft_bounds2 = pg.InfiniteLine(pen=self.pen_fftbounds, movable=False)
        self.tdo_inverse_field.addItem(self.fft_bounds1)
        self.tdo_inverse_field.addItem(self.fft_bounds2)
        self._plots_list.add(self.tdo_inverse_field)

    def create_fft_plot(self):
        # vs field
        self.fft = pg.PlotWidget(title="Fourier transform")
        self.fft.setLabel("bottom", "B-frequency (T)")
        self.fft.setLabel("left", "magnitude")
        self.fft.getPlotItem().addLegend()
        self._plots_list.add(self.fft)

    def customize_axis(self):
        """Set a right axis and show the grid, for all the plots."""
        for plot in self._plots_list:
            plot.showAxis("right")
            plot.getAxis("right").setTicks([])
            plot.showGrid(y=True)

    def init_plot_style(self):
        """Set up PyQtGraph line styles."""
        w0 = 1
        w1 = 2
        self.pen_field = pg.mkPen("#c7c7c7", width=w0)
        self.pen_bup = pg.mkPen("#2ca02cff", width=w0)
        self.pen_bdown = pg.mkPen("#d62728ff", width=w0)
        self.pen_tdo = pg.mkPen("#fffc5c80", width=w0)
        self.pen_fitbup = pg.mkPen("#2ca02cb1", width=w0)
        self.pen_fitdown = pg.mkPen("#d62728b1", width=w0)
        self.pen_fitbounds = pg.mkPen("#7477ed99", width=w1)
        self.pen_fftbounds = pg.mkPen("#eddd86ff", width=w1)

    @pyqtSlot()
    def roi1_changed(self):
        """Update ROI in the TDO detrended panel when the ROI in the TDO panel moved."""
        roi1 = self.roi.getRegion()  # TDO panel
        # Show the select fit range in the TDO detrended panel
        self.fit_bounds1.setPos(roi1[0])
        self.fit_bounds2.setPos(roi1[1])
        self.sig_roi1_changed.emit()
        if not self._sync_roi:
            return

        roi2 = self.roi2.getRegion()  # TDO detrend panel

        if roi1 != roi2:
            self.roi2.setRegion(self.roi.getRegion())

    @pyqtSlot()
    def roi2_changed(self):
        """Update ROI in the TDO panel when the ROI in the TDO detrended panel moved."""
        roi2 = self.roi2.getRegion()  # TDO detrend panel
        # Show the FFT range in the TDO 1/B panel
        self.fft_bounds1.setPos(1 / roi2[0])
        self.fft_bounds2.setPos(1 / roi2[1])
        self.sig_roi2_changed.emit()
        if not self._sync_roi:
            return

        roi1 = self.roi.getRegion()  # TDO panel

        if roi1 != roi2:
            self.roi.setRegion(self.roi2.getRegion())

    def enable_rois(self):
        """Enable moving ROIs."""
        self.roi.setMovable(True)
        self.roi2.setMovable(True)

    def disable_rois(self):
        """Disable moving ROIs."""
        self.roi.setMovable(False)
        self.roi2.setMovable(False)

    @pyqtSlot(int)
    def move_plots(self, index: int):
        """Move plots from one tab to another, because a widget can't be copied."""
        if index == 0:
            # First tab
            self.tab2_top_splitter.replaceWidget(0, self._v2field)
            self.tab2_top_splitter.replaceWidget(1, self._v2sig_field)
            self.tab1_top_splitter.replaceWidget(1, self.field)
            self.tab1_main_splitter.replaceWidget(1, self.sig_field)
        elif index == 1:
            # Second tab
            self.tab1_top_splitter.replaceWidget(1, self._v1field)
            self.tab1_main_splitter.replaceWidget(1, self._v1sig_field)
            self.tab2_top_splitter.replaceWidget(0, self.field)
            self.tab2_top_splitter.replaceWidget(1, self.sig_field)



"""
The `pyuson` subpackage, a library for ultrasound echoes experiments.

Provides
--------
- `EchoConfig` : The `Config` object specific to ultrasound experiments.
- `EchoProcessor` : the `Processor` object specific to ultrasound experiments.
- `config_models` : the `pydantic` models to define the configuration requirements.
- `gui` : a graphical application. It needs to be imported explicitely.
"""

from . import config_models
from ._config import EchoConfig
from ._echoprocessor import EchoProcessor

__all__ = ["EchoConfig", "EchoProcessor", "config_models"]



"""Logging utilities."""

import logging
import sys
import warnings
from logging.handlers import RotatingFileHandler
from pathlib import Path
from types import TracebackType
from typing import Literal

from rich.logging import RichHandler

__all__ = ["configure_logger"]

# Setup logging directory
APP_DIR = Path.home() / ".pymagnetos"
APP_DIR.mkdir(parents=True, exist_ok=True)

APP_NAME = "pymagnetos"
LOGGING_FORMAT = "{asctime}.{msecs:3g} [{levelname}] : {message}"
CONSOLE_HANDLER = RichHandler(log_time_format="%H:%M:%S.%f")
FILE_FORMATTER = logging.Formatter(
    LOGGING_FORMAT, style="{", datefmt="%Y-%m-%d %H:%M:%S"
)


def configure_logger(
    logger: logging.Logger, filename: str, log_level: str = "INFO"
) -> None:
    """
    Configure logger handlers.

    Parameters
    ----------
    filename : str
        Name of the file in which log will be written to. The file will be created under
        $HOME/.pymagnetos/.
    log_level : str, optional
        Logging level to choose from ("DEBUG", "INFO", "WARNING", "ERROR"). Default is
        "INFO".
    """
    # Define the file handler
    log_file = APP_DIR / filename
    file_handler = RotatingFileHandler(
        log_file, mode="a", encoding="utf-8", maxBytes=int(1e6), backupCount=1
    )
    file_handler.setLevel(log_level)
    file_handler.setFormatter(FILE_FORMATTER)

    # Add handlers and set log levels
    remove_file_handlers(logger, log_file, mode="same")
    CONSOLE_HANDLER.setLevel(log_level)
    logger.addHandler(CONSOLE_HANDLER)
    logger.addHandler(file_handler)
    logger.setLevel(log_level)

    # Get the core logger and add the same handlers
    core_logger = logging.getLogger(APP_NAME + ".core")
    remove_file_handlers(core_logger, str(log_file), mode="all")
    core_logger.addHandler(CONSOLE_HANDLER)
    core_logger.addHandler(file_handler)
    core_logger.setLevel(log_level)

    # Set the logger to log exceptions
    log_uncaugth_exceptions(logger)

    logger.info(f"Logging to {log_file}")


def remove_file_handlers(
    logger: logging.Logger,
    file: str | Path,
    mode: Literal["same", "all"] = "same",
    warn: bool = True,
) -> None:
    """
    Remove file handlers from a logger.

    In mode "same", only handlers that log to the same file will be removed, in mode
    "all", all file handlers are removed.

    Parameters
    ----------
    logger : logging.Logger
        Logger instance.
    file : str or Path
        Path to the log file.
    mode : {"same", "all"}, optional
        "same" : remove the handler only if it logs to the same file (default).
        "all": remove all file handlers.
    warn : bool, optional
        Issue a warning when removing a handler set to a different file. Default is
        True.
    """
    file = str(file)
    for idx in range(len(logger.handlers)):
        handler = logger.handlers[idx]
        if isinstance(handler, logging.FileHandler):
            if handler.baseFilename == file:
                logger.removeHandler(handler)
            else:
                if mode == "all":
                    if warn:
                        warnings.warn(
                            f"The logger was writing to {handler.baseFilename}, "
                            "it will stop"
                        )
                    logger.removeHandler(handler)


def log_uncaugth_exceptions(logger: logging.Logger) -> None:
    """Make uncaugth exceptions to be logged by the logger."""

    def handle_exception(
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType | None,
    ) -> None:
        if not issubclass(exc_type, KeyboardInterrupt):
            logger.critical(
                "Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback)
            )

        sys.__excepthook__(exc_type, exc_value, exc_traceback)

    sys.excepthook = handle_exception



"""The grid with main action buttons."""

from PyQt6 import QtWidgets
from PyQt6.QtCore import pyqtSignal


class ButtonsWidget(QtWidgets.QWidget):
    sig_load = pyqtSignal()
    sig_extract = pyqtSignal()
    sig_analyse = pyqtSignal()
    sig_tdocsv = pyqtSignal()
    sig_rescsv = pyqtSignal()
    sig_save_nexus = pyqtSignal()

    def __init__(self) -> None:
        super().__init__()

        grid = QtWidgets.QGridLayout()

        self.button_load = QtWidgets.QPushButton("Load data", self)
        self.button_extract = QtWidgets.QPushButton("Extract TDO", self)
        self.button_analyse = QtWidgets.QPushButton("Oscillations analysis", self)
        self.button_tdocsv = QtWidgets.QPushButton("Export TDO as CSV", self)
        self.button_rescsv = QtWidgets.QPushButton("Export results as CSV", self)
        self.button_save_nexus = QtWidgets.QPushButton("Save as NeXus", self)

        self.disable_buttons()

        self.connect_buttons()

        grid.addWidget(self.button_load, 0, 0)
        grid.addWidget(self.button_extract, 0, 1)
        grid.addWidget(self.button_analyse, 1, 0, 1, 2)

        grid_save = QtWidgets.QHBoxLayout()
        grid_save.addWidget(self.button_tdocsv)
        grid_save.addWidget(self.button_rescsv)
        grid_save.addWidget(self.button_save_nexus)

        grid.addLayout(grid_save, 2, 0, 1, 2)

        self.setLayout(grid)

    def connect_buttons(self) -> None:
        self.button_load.clicked.connect(self.sig_load.emit)
        self.button_extract.clicked.connect(self.sig_extract.emit)
        self.button_analyse.clicked.connect(self.sig_analyse.emit)
        self.button_tdocsv.clicked.connect(self.sig_tdocsv.emit)
        self.button_rescsv.clicked.connect(self.sig_rescsv.emit)
        self.button_save_nexus.clicked.connect(self.sig_save_nexus.emit)

    def disable_buttons(self) -> None:
        self.button_load.setEnabled(False)
        self.button_extract.setEnabled(False)
        self.button_analyse.setEnabled(False)
        self.button_tdocsv.setEnabled(False)
        self.button_rescsv.setEnabled(False)
        self.button_save_nexus.setEnabled(False)

    def enable_buttons(self) -> None:
        self.button_load.setEnabled(False)
        self.button_extract.setEnabled(True)
        self.button_analyse.setEnabled(True)
        self.button_tdocsv.setEnabled(True)
        self.button_rescsv.setEnabled(True)
        self.button_save_nexus.setEnabled(True)



"""
Generic application oriented toward a configuration > analysis > display workflow.

It consists of a Configuration Panel on the left, a graphs area on the right, some
buttons on the bottom right corner and the application log on the bottom right.
The Configuration Panel is a tab, other tabs are : a file browser and a batch processing
tab.

The BaseMainWindow should be sub-classed to customize buttons and plots and connect the
signals to a worker embeding a Processor object. See example implementations in the
`pyuson` or `pytdo` subpackages.
"""

import logging
import re
from collections.abc import Sequence
from pathlib import Path

import pyqtgraph as pg
from PyQt6 import QtGui, QtWidgets
from PyQt6.QtCore import Qt, QThread
from PyQt6.QtCore import pyqtSignal as Signal
from PyQt6.QtCore import pyqtSlot as Slot

from .widgets import TextLoggerWidget
from .worker import BaseWorker

REGEXP_EXPID_SEPARATORS = r"[_-]"
ALLOWED_FORMAT = ("toml", "json", "nx5", "nxs", "h5", "hdf5")


class BaseMainWindow(QtWidgets.QMainWindow):
    """
    A generic graphical user interface application.

    The `MainWindow` class defines the main thread with the front-end user interface.
    It creates the main general layout and include some generic functions related to
    standard analysis processes : generate a PyQtGraph ParameterTree, connect to a
    worker in its own thread, load and plot a magnetic field...

    The main window consists of the following :
    - A left panel with three tabs, the Configuration tab, a file browser and a Batch
    processing tab,
    - A right panel with all the plots. There can be any number of graphs, in any number
    of tabs,
    - A bottom-left box with the main action button,
    - A bottom-right box with the log.

    It is built with pre-defined widgets that must implement a specific interface :
    One should subclass this `BaseMainWindow` class and register those widgets as
    private attributes before calling `super().__init__()`. The widgets' class must be
    registered, not an instance of such, e.g. do `ConfigurationWidget`, not
    `ConfigurationWidget()`. The mandatory widgets to implement and register are :

    - _type_wconfiguration : a configuration widget for the parameter tree
    - _param_content : a simple class that provides the ParameterTree content,
    - _type_wfiles : a file browser widget,
    - _type_wbatch : a batch-processing widget,
    - _type_wgraphs : a graph widget that holds all the plots,
    - _type_wbuttons : a widget with the main action buttons,
    - _type_worker : a worker object.

    Each of thos classes are instantiated here with the `_type_` removed, e.g.
    `_type_wgraphs` is intantiated as `wgraphs`.

    Each of those components exposes signals, to which the main thread connects its own
    methods, instead of connecting directly to the widgets (buttons, checkboxes and the
    like). Those signals are named with a trailing `sig_` so that makes it easy to know
    exactly what should be connected in the main thread.

    When loading a configuration (or a NeXus) file, the main thread creates a worker
    thread. The latter instantiates an Processor object that perform the actual
    analysis.

    Signals
    -------
    All signals emit in their corresponding methods and are connected to the
    corresponding methods in the worker thread. Any number of supplementary signals can
    be implemented in subclasses.

    sig_worker_load
    sig_worker_preprocess
    sig_worker_align
    sig_worker_analysis
    sig_worker_batch
    sig_worker_save_nexus
    """

    sig_worker_load: Signal = Signal()
    sig_worker_compute_field: Signal = Signal()
    sig_worker_preprocess: Signal = Signal()
    sig_worker_align: Signal = Signal()
    sig_worker_analysis: Signal = Signal()
    sig_worker_batch: Signal = Signal(tuple)
    sig_worker_save_nexus: Signal = Signal(str)

    _type_wbatch: type
    _type_wbuttons: type
    _type_wconfiguration: type
    _type_wfiles: type
    _type_wgraphs: type
    _param_content: type
    _type_worker: type[BaseWorker]

    _do_reload_expid: bool
    _do_load_data: bool
    _do_preprocess: bool
    _do_analysis: bool
    _ignore_tree_change: bool
    worker: BaseWorker

    def __init__(self):
        super().__init__()

        # Set flags
        self._do_reload_expid = True  # reload the experiment ID dataset
        self._do_load_data = True  # load raw data
        self._do_preprocess = True  # perform preprocessing
        self._do_analysis = True  # perform analysis
        self._ignore_tree_change = False  # ignore parameter being changed in the tree
        self.ind_bup = slice(0)  # magnetic field up indices
        self.ind_bdown = slice(0)  # magnetic field down indices

        # Initialize components
        self.init_parameter_tree()
        self.init_files()
        self.init_batch_processing()
        self.init_buttons()
        self.init_plots()
        self.init_log()
        self.init_layout()
        self.setCentralWidget(self.splitter0)

        # Setup drag&drop
        self.setAcceptDrops(True)

        # Show window
        self.show()

    def init_parameter_tree(self):
        """
        Create the "Configuration" tab.

        It holds the PyQtGraph parameter tree. Each parameter is synchronised with the
        `Config` object of the Processor worker.
        """
        self.wconfiguration = self._type_wconfiguration(self._param_content)

        # Connect
        self.wconfiguration.sig_file_changed.connect(self.load_file)
        self.wconfiguration.sig_expid_changed.connect(self.reload_expid)
        self.wconfiguration.sig_autoload_changed.connect(
            self.update_autoload_from_config
        )
        self.wconfiguration.sig_reload_config.connect(self.load_file)
        self.wconfiguration.sig_save_config.connect(self.save_config)

        # Watch changes from the parameter tree to update the worker configuration
        self.wconfiguration.sig_parameter_changed.connect(
            self.set_worker_config_from_tree
        )
        # Connect additional signals from the configuration panel
        self.connect_parameter_tree()

    def init_files(self):
        """
        Create the "Files" tab.

        It lists files in a directory, one can double-click a file to load either a
        configuration file or change the experiment ID.
        """
        self.wfiles = self._type_wfiles()

        # Connect
        self.wfiles.sig_file_selected.connect(self.select_file_in_browser)
        self.wfiles.sig_checkbox_changed.connect(self.update_autoload_from_files)

    def init_batch_processing(self):
        """
        Create the "Batch processing tab.

        It's multi-panel file picker, where the user can queue files to be
        batch-processed with the same settings.
        """
        self.wbatch = self._type_wbatch()

        # Connect
        self.wbatch.sig_batch_process.connect(self.batch_process)

    def init_buttons(self):
        """Create the main action buttons."""
        self.wbuttons = self._type_wbuttons()

        # Connect
        self.wbuttons.sig_load.connect(self.load_data)
        self.wbuttons.sig_save_nexus.connect(self.save_nexus)

    def init_plots(self):
        """Create the graph area that holds all the plots."""
        self.wgraphs = self._type_wgraphs()

    def init_log(self):
        """Initiliaze the log, calling the `_init_log()` method."""
        raise NotImplementedError(
            "Subclasses must implement this method, using the "
            "_init_log(program_name, log_level) method"
        )

    def _init_log(self, program_name: str, log_level: str = "info"):
        """
        Create the widget where the log will be printed.

        The logger is shared with the package-level logger so stream is also printed in
        stdout and the log file.
        """
        log_handler = TextLoggerWidget(self)

        # Set up the formatter
        log_handler.setFormatter(
            logging.Formatter(
                "{asctime}.{msecs:3g} [{levelname}] : {message}",
                style="{",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        )

        # Set up the logger
        self.logger = logging.getLogger(program_name)
        self.logger.addHandler(log_handler)
        self.logger.setLevel(log_level)

        # Store the logger widget (QPlainTextEdit)
        self.wlog = log_handler.widget

    def init_layout(self):
        """
        Set up the main layout.

        It is composed of a 4 panels seperated with resizable splitters.
        Panel 1 (top left) : Tabs with Configuration, Files and Batch processing
        Panel 2 (bottom left) : Action buttons
        Panel 3 (top right) : Graphs
        Panel 4 (bottom right) : Log
        """
        self.splitter0 = QtWidgets.QSplitter(self)
        splitter1 = QtWidgets.QSplitter(Qt.Orientation.Vertical, self)
        splitter2 = QtWidgets.QSplitter(Qt.Orientation.Vertical, self)

        self.config_tabs = QtWidgets.QTabWidget(self)
        self.config_tabs.addTab(self.wconfiguration, "Configuration")
        self.config_tabs.addTab(self.wfiles, "Files")
        self.config_tabs.addTab(self.wbatch, "Batch processing")

        splitter1.addWidget(self.config_tabs)
        splitter1.addWidget(self.wbuttons)

        splitter2.addWidget(self.wgraphs)
        splitter2.addWidget(self.wlog)

        self.splitter0.addWidget(splitter1)
        self.splitter0.addWidget(splitter2)

    @Slot()
    def load_file(self):
        """
        Load a configuration or NeXus file and start a worker.

        Callback for when the "File" parameter changes, or a file is drag & dropped in
        the window.
        """
        if not Path(self.wconfiguration.files_parameters["file"]).is_file:
            self.logger.error("[GUI] Input file does not exist")
            return
        try:
            self.create_worker()
        except Exception as e:
            self.logger.error(f"There was an error creating worker : {e}")
            self.wconfiguration.button_reload_config.setEnabled(True)
        try:
            self.plot_field()
            self.get_up_down_indices()
        except Exception as e:
            self.logger.error(f"There was an error plotting field : {e}")
            self.wconfiguration.button_reload_data.setEnabled(True)

        # Set experiment ID read from the configuration, without reloading it
        self._do_reload_expid = False
        self.wconfiguration.files_parameters["expid"] = self.worker.proc.expid
        self._do_reload_expid = True

        # Update file explorer directory
        self.wfiles.file_browser.set_directory(self.worker.proc.cfg.data_directory)
        self.wbatch.current_directory = self.worker.proc.cfg.data_directory
        self.wbatch.prefix = re.split(REGEXP_EXPID_SEPARATORS, self.worker.proc.expid)[
            0
        ]

        # Enable buttons
        self.wconfiguration.button_reload_data.setEnabled(True)
        self.wconfiguration.button_save_config.setEnabled(True)

        # Load data directly if autoload is on
        if (
            self.wconfiguration.files_parameters["autoload"]
            and not self.worker.is_data_loaded
        ):
            self.load_data()

    def create_worker(self):
        """Create the Processor object in its own worker thread."""
        # Delete previous instance if any
        if hasattr(self, "worker_thread"):
            self.reset()

        # Create thread
        self.worker_thread = QThread()
        # Create worker
        self.worker = self._type_worker(self.wconfiguration.files_parameters["file"])

        # Move worker to thread
        self.worker.moveToThread(self.worker_thread)

        # Start the thread
        self.worker_thread.start()

        # Read the configuration
        self.set_tree_from_worker()

        # Connect signals and slots
        self.connect_worker()

        self.logger.info("[GUI] Processor thread created and connected")

    def set_tree_from_worker(self):
        """
        Set the parameters in the parameter tree from the worker configuration.

        Loop through all parameter in the "parameters" and "settings" section and set
        them from the `Config` object of the Processor object.
        """
        # In the Parameters section
        for p in self.wconfiguration.param_parameters:
            self.set_param_from_worker(p.name(), context="parameters")
        # In the Settings section
        for p in self.wconfiguration.settings_parameters:
            self.set_param_from_worker(p.name(), context="settings")

    def set_param_from_worker(self, param_name: str, context: str):
        """
        Set a parameter read from the `Config` object of the Processor worker.

        Parameters
        ----------
        param_name : str
            Name of the parameter in the tree and `Config`.
        context : {"parameters", "settings"}
            Configuration file section.
        """
        match context:
            case "parameters":
                if hasattr(self.worker.proc.cfg.parameters, param_name):
                    self.wconfiguration.param_parameters[param_name] = getattr(
                        self.worker.proc.cfg.parameters, param_name
                    )
            case "settings":
                if hasattr(self.worker.proc.cfg.settings, param_name):
                    config_value = getattr(self.worker.proc.cfg.settings, param_name)
                    if param_name in self.wconfiguration.parameters_to_parse:
                        # special case that needs to be converted to string
                        config_value = self.wconfiguration.get_text_from_numbers(
                            config_value
                        )
                    self.wconfiguration.settings_parameters[param_name] = config_value
            case _:
                raise ValueError(f"Error : unknown configuration section : {context}")

    def connect_worker(self):
        """Connect signals from the main thread to tasks in the worker."""
        # Load data
        self.sig_worker_load.connect(self.worker.load_data)
        self.worker.sig_load_data_finished.connect(self.load_data_finished)

        # Compute field
        self.sig_worker_compute_field.connect(self.worker.compute_field)
        self.worker.sig_compute_field_finished.connect(self.compute_field_finished)

        # Preprocessing
        self.sig_worker_preprocess.connect(self.worker.preprocess)
        self.worker.sig_preprocess_finished.connect(self.preprocess_finished)

        # Align field
        self.sig_worker_align.connect(self.worker.align_field)
        self.worker.sig_align_finished.connect(self.align_field_finished)

        # Analysis
        self.sig_worker_analysis.connect(self.worker.run_analysis)
        self.worker.sig_analysis_finished.connect(self.run_analysis_finished)

        # Batch process
        self.sig_worker_batch.connect(self.worker.batch_process)
        self.worker.sig_batch_finished.connect(self.batch_process_finished)

        # Save as NeXus
        self.sig_worker_save_nexus.connect(self.worker.save_as_nexus)
        self.worker.sig_save_nexus_finished.connect(self.save_nexus_finished)

    def connect_parameter_tree(self) -> None:
        """
        Connect signals from the Configuration widget.

        Changes in the parameter tree should be handled in the `run_after_tree_change()`
        method instead.
        """
        pass

    @Slot(str, str, object)
    def set_worker_config_from_tree(
        self,
        context: str,
        param_name: str,
        value: float | int | list | bool | str,
    ):
        """
        Read parameter from the Parameter Tree and set its sibling in the Config object.

        Callback for any change in the "Parameters" or "Settings" sections of the
        parameter tree, the arguments are passed with the signal.

        After the parameter is changed, the `run_after_tree_change()` method is called,
        passing the context, parameter name and value.

        Note : if the `_ignore_tree_change` attribute is False, this method does not do
        anything.

        Parameters
        ----------
        context : {"parameters", "settings"}
            Define where the key is stored in the Config object (corresponding to the
            configuration file section).
        param_name : str
            Name of the parameter in the Parameter Tree.
        value : Any
            The parameter value.
        """
        if self._ignore_tree_change:
            return
        match context:
            case "parameters":
                setattr(self.worker.proc.cfg.parameters, param_name, value)
            case "settings":
                setattr(self.worker.proc.cfg.settings, param_name, value)
            case _:
                raise ValueError(f"Unknown configuration section : {context}")

        self.run_after_tree_change(context, param_name, value)

    def run_after_tree_change(
        self, context: str, param_name: str, value: float | int | list | bool | str
    ) -> None:
        """Run additional actions after a parameter change in the parameter tree."""
        pass

    @Slot()
    def reload_expid(self):
        """
        Reload experiment ID, resetting the data via the worker.

        `expid` is a property of the Processor object, when it is changed, it triggers a
        reinitializtion of the object.

        If data autoloading is enabled, the data is reloaded here.

        Callback for a change of the experiment ID in the "Files" section of the
        parameter tree and of the `Reload` button.
        """
        if not self.check_config_loaded():
            self.logger.warning("[GUI] No configuration file loaded")
            return
        if not self._do_reload_expid:
            # The experiment ID is being changed programatically, do not update
            return

        # Infer experiment ID
        expid = self.infer_expid(self.wconfiguration.files_parameters["expid"])
        # Set it, without reloading since that's we're doing
        self._do_reload_expid = False
        self.wconfiguration.files_parameters["expid"] = expid
        self._do_reload_expid = True

        self.logger.info(
            f"[GUI] Reloading experiment ID, from {self.worker.proc.expid} to {expid}"
        )

        # Set the new experiment ID in the worker
        self.worker.proc.expid = expid
        self.worker.is_data_loaded = False

        # Prepare
        self.disable_buttons()
        self.wgraphs.clear_all_plots()
        self.plot_field()
        self.wconfiguration.button_reload_data.setEnabled(True)
        self.wconfiguration.button_save_config.setEnabled(True)
        self.wbuttons.button_load.setEnabled(True)

        if self.wconfiguration.files_parameters["autoload"]:
            # Autoload data
            self.load_data()

    @Slot()
    def save_config(self):
        """
        Save current configuration as a TOML file.

        A default file name is generated and a file picker dialog box is openned for the
        user to choose an output file.
        """
        if not self.check_config_loaded():
            return

        cfg_file = Path(self.wconfiguration.files_parameters["file"])
        default_fname = cfg_file.with_stem(cfg_file.stem + "-2")
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Save configuration file as...",
            directory=str(default_fname),
            filter="TOML files (*.toml), JSON files (*.json)",
            options=QtWidgets.QFileDialog.Option.DontConfirmOverwrite,
        )
        self.worker.proc.cfg.write(fname)

    @Slot()
    def load_data(self):
        """
        Load data.

        Send the signal to the worker to load data. Callback for the "Load data" button.
        It's also used whenever a new dataset is loaded, if data auto-loading is
        enabled.
        """
        if not self.check_config_loaded():
            self.logger.warning("[GUI] No configuration file loaded")
            return
        if not self._do_load_data:
            return

        self.disable_buttons()
        self.sig_worker_load.emit()

    @Slot()
    def load_data_finished(self):
        """Callback for when the worker has finished loading data."""
        self.enable_buttons()

    @Slot()
    def compute_field(self) -> None:
        """Re-compute magnetic field."""
        if not self.check_config_loaded():
            return
        self.disable_buttons()
        self.sig_worker_compute_field.emit()

    @Slot()
    def compute_field_finished(self):
        """Callback for when the worker has finished computing the magnetic field."""
        self.plot_field()
        self.enable_buttons()

    @Slot()
    def preprocess(self):
        """Run the preprocessing steps."""
        if not self.check_data_loaded():
            return
        if not self._do_preprocess:
            return

        self.disable_buttons()
        self.sig_worker_preprocess.emit()

    @Slot()
    def preprocess_finished(self):
        """Callback for when the preprocessing steps are finished."""
        self.enable_buttons()

    @Slot()
    def align_field(self):
        """Align the magnetic field on the data."""
        if not self.check_data_loaded():
            return

        self.disable_buttons()
        self.sig_worker_align.emit()

    @Slot()
    def align_field_finished(self):
        """
        Re-plot magnetic field with its new time vector.

        Callback for when the worker has finished loading data. This is re-triggered
        everytime the time vector is susceptible to change (subsampling, ...).
        """
        # Checks
        if not self.check_field_aligned():
            return

        # Re-plot magnetic field
        self.plot_field()

        # Get indexer for field up and field down
        self.get_up_down_indices()
        self.enable_buttons()

    @Slot()
    def run_analysis(self):
        """Run the analysis."""
        if not self.check_data_loaded():
            return
        if not self._do_analysis:
            return

        self.disable_buttons()
        self.sig_worker_analysis.emit()

    @Slot()
    def run_analysis_finished(self):
        """Callback for when the analysis is finished."""
        self.enable_buttons()

    @Slot()
    def save_nexus(self):
        """
        Save everything as a NeXus file.

        The output file is self-contained, it saves all parameters, echo indices, raw
        and processed data.

        Generate a default file name and open a file picker dialog for the user to
        choose an output file, and send the signal to the worker to save the results as
        NeXus.
        Callback for the "Save as NeXus" button.
        """
        if not self.check_data_loaded():
            return

        self.disable_buttons()

        default_fname = self.worker.proc.get_nexus_filename()
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Save NeXus file as...",
            default_fname,
            "NeXus files (*.hdf5, *.h5, *.nxs, *.nx5)",
        )

        if fname:
            self.sig_worker_save_nexus.emit(fname)
        else:
            self.logger.error("[GUI] Invalid output file name for NeXus file")
            self.save_nexus_finished()

    @Slot()
    def save_nexus_finished(self):
        """Callback for when the worker has finished saving as NeXus."""
        self.enable_buttons()

    @Slot()
    def batch_process(self):
        """
        Run batch processing on selected files.

        List files in the queue, determine corresponding experiment IDs, create a
        progress bar and send the signal to the worker to perform the batch-processing.
        Callback for the "Batch process" button in the "Batch processing" tab.
        """
        if not self.check_config_loaded():
            self.logger.warning("[GUI] No configuration was loaded")
            return

        # List files in the queue
        self.batch_files = self.wbatch.get_files_to_process()

        if len(self.batch_files) == 0:
            self.logger.warning("[GUI] No items in batch processing list")
            return

        # Infer experiments ID from the file names
        self.batch_expids = []
        for file in self.batch_files:
            self.batch_expids.append(self.infer_expid(file))

    @Slot(int)
    def batch_process_step_finished(self, idx: int):
        """Track progress of the batch processing."""
        # Move file in the "Done" list
        self.wbatch.move_to_done(self.batch_files[idx])

    @Slot()
    def batch_process_finished(self):
        """
        Cleanup after batch-processing.

        Callback for when the worker has finished the batch-processing.
        """
        self.worker.sig_batch_progress.disconnect()

        # Set the last experiment ID in the parameter tree, without reloading since it
        # is already the currently loaded dataset
        self._do_reload_expid = False
        self.wconfiguration.files_parameters["expid"] = self.batch_expids[-1]
        self._do_reload_expid = True
        self.batch_files = []
        self.batch_expids = []

    @Slot(bool, str)
    def select_file_in_browser(self, is_toml: bool, filepath: str):
        """
        Select a file and set it as a configuration file or a new experiment ID.

        Callback for when the user double-click on a file in the "Files" tab.
        """
        if is_toml:
            self.wconfiguration.files_parameters["file"] = filepath
        else:
            self.wconfiguration.files_parameters["expid"] = filepath

    def infer_expid(self, expid: str) -> str:
        """
        Determine experiment ID from a file name or path.

        If the input is an existing file, it uses the base file name without extension,
        removes "-pickup" and, for WFM files, the channels number, to build the
        experiment ID. Otherwise, it is used as is.
        """
        file_path = Path(expid)

        if file_path.is_file():
            # Input is a file, get the corresponding experiment ID
            if expid.lower().endswith(".wfm"):
                # Remove the _chX bit for WFM files
                res = re.sub(r"_ch\d+\.wfm$", ".wfm", expid, flags=re.IGNORECASE)
                file_path = Path(res)

            infered_expid = file_path.stem.replace("-pickup", "")

            return infered_expid
        else:
            # Input experiment ID is a proper experiment ID, return as is
            return expid

    def _handle_roi_changed(
        self, roi: pg.LinearRegionItem, settings_name: str, precision: int = 6
    ) -> None:
        """
        Define what to do when a PyQtGraph ROI is changed.

        This is a helper function to handle a ROI being moved in a graph. In the
        Parameter tree, 2-elements lists are displayed as text, which requires parsing.
        To make sure to use the actual ROI values and not the parsed value (which has a
        given precision), the ROI is set directly in the worker. The values in the tree
        is then set with a reasonable precision to not clutter the interface.

        Subclasses need to implement their own `handle_roi_change()` slot method
        connected to a ROI changed event. It should get the relevant ROI, and use the
        present helper to update the setting in both the worker and the tree.
        """
        # Get time range
        xmin, xmax = roi.getRegion()

        # Check the ROI was changed since it was created
        if (xmin, xmax) == (0, 1):
            return

        # Update parameter in the worker directly to ensure real values are used without
        # parsing. This is because in the tree the range is displayed as a string.
        if self.check_config_loaded():
            setattr(self.worker.proc.cfg.settings, settings_name, [xmin, xmax])

        # The displayed numbers have 16 digits, so trim it, without updating the worker
        self._ignore_tree_change = True
        self.wconfiguration.settings_parameters[settings_name] = (
            self.wconfiguration.get_text_from_numbers([xmin, xmax], precision=precision)
        )
        self._ignore_tree_change = False

    def _update_roi(self, roi: pg.LinearRegionItem, new_region: Sequence[float | int]):
        """
        Update the ROI in the graph from "Analysis time window" in the parameter tree.

        Subclasses need to implement their own `update_roi()` method with the relevant
        ROI.
        """
        # Sanity checks
        if len(new_region) != 2:
            self.logger.error(f"[GUI] Invalid analysis window : {new_region}")
            return
        elif new_region[0] >= new_region[1]:
            self.logger.error(f"[GUI] Invalid analysis window : {new_region}")
            return

        # Update ROI
        roi.setRegion(new_region)

    def plot_field(self):
        """
        Display magnetic field versus time.

        The pickup coil voltage and the magnetic field are shown. If it is aligned with
        the experiment time vector, a cross-hair tracks the frame number on mouse hover.
        """
        if not self.worker.proc.get_data_processed("magfield", checkonly=True):
            self.worker.proc.compute_field()

            if not self.worker.proc.get_data_processed("magfield", checkonly=True):
                # Still no field to plot, maybe there was no pickup and data was not
                # loaded so the pickup signal could be simulated
                self.logger.warning("[GUI] Field was not computed, not plotting")
                return

        # B(t)
        self.wgraphs.field.clearPlots()
        self.wgraphs.field.plot(
            self.worker.proc.get_data_processed("magfield_time"),
            self.worker.proc.get_data_processed("magfield"),
            pen=self.wgraphs.pen_field,
        )
        self.wgraphs.field.setTitle(
            "Magnetic field (max. field : "
            f"{self.worker.proc.get_data_processed('magfield').max():2.2f}T)"
        )

        # dB/dt (pickup)
        self.wgraphs.dfield.clearPlots()
        if self.worker.proc.get_data_raw(
            "pickup_time", checkonly=True
        ) and self.worker.proc.get_data_raw("pickup", checkonly=True):
            self.wgraphs.dfield.plot(
                self.worker.proc.get_data_raw("pickup_time"),
                self.worker.proc.get_data_raw("pickup"),
                pen=self.wgraphs.pen_field,
            )

    def plot_var_time(self, pgplot, varname: str, mult: float = 1):
        """
        Plot `varname` versus experiment time.

        Used to plot attenuation and phase-shift versus time.
        """
        pgplot.clearPlots()
        pgplot.plot(
            self.worker.proc.get_data_processed("time_exp"),
            self.worker.proc.get_data_serie(varname) * mult,
        )

    def plot_var_field(self, pgplot, varname: str, mult: float = 1):
        """Plot `varname` versus magnetic field.

        Field rise and decay are distinguished with a different color. Corresponding
        indices should be already available from the `ind_bdown` and `ind_bup`
        attributes, that are created in the `align_field_finished()` method.
        """
        pgplot.clearPlots()
        pgplot.plot(
            self.worker.proc.get_data_processed("magfield")[self.ind_bdown],
            self.worker.proc.get_data_serie(varname)[self.ind_bdown] * mult,
            pen=self.wgraphs.pen_bdown,
            name="B down",
        )
        pgplot.plot(
            self.worker.proc.get_data_processed("magfield")[self.ind_bup],
            self.worker.proc.get_data_serie(varname)[self.ind_bup] * mult,
            pen=self.wgraphs.pen_bup,
            name="B up",
        )

    def disable_buttons(self):
        """Disable all buttons and ROIs."""
        self.wconfiguration.disable_buttons()
        self.wbuttons.disable_buttons()
        self.wbatch.disable_buttons()
        self.wgraphs.disable_rois()

    def enable_buttons(self):
        """Enable all buttons and ROIs."""
        self.wconfiguration.enable_buttons()
        self.wbuttons.enable_buttons()
        self.wbatch.enable_buttons()
        self.wgraphs.enable_rois()

    @Slot()
    def update_autoload_from_files(self):
        """Update the autoload checkbox in the Configuration when toggled in Files."""
        self.wconfiguration.files_parameters["autoload"] = (
            self.wfiles.checkbox_autoload_data.isChecked()
        )

    @Slot()
    def update_autoload_from_config(self):
        """Update the autoload checkbox in Files when toggled in the Configuration."""
        self.wfiles.checkbox_autoload_data.setChecked(
            self.wconfiguration.files_parameters["autoload"]
        )

    def get_up_down_indices(self) -> None:
        """Store the magnetic field increase and decrease indices."""
        self.ind_bup = self.worker.proc.inds_b_inc  # increasing B
        self.ind_bdown = self.worker.proc.inds_b_dec  # decreasing B

    def check_config_loaded(self) -> bool:
        """Check if a worker was initialized."""
        if hasattr(self, "worker"):
            return True
        else:
            return False

    def check_data_loaded(self) -> bool:
        """Check if data was loaded."""
        if not self.check_config_loaded():
            return False
        else:
            if not self.worker.is_data_loaded:
                return False
            else:
                return True

    def check_field_aligned(self) -> bool:
        """Check if the magnetic field is aligned on the experiment time vector."""
        if not self.check_config_loaded():
            return False

        return self.worker.proc._check_field_aligned()

    def reset(self):
        """Quit and delete worker and thread, resetting plots and parameters."""
        # Quit thread
        if hasattr(self, "worker_thread"):
            self.worker_thread.quit()
            self.worker_thread.deleteLater()
        if hasattr(self, "worker"):
            self.worker.deleteLater()

        # Clear plots
        self.wgraphs.clear_all_plots()
        # Disable buttons
        self.disable_buttons()

    def dragEnterEvent(self, a0: QtGui.QDropEvent | None = None) -> None:
        """Handle drag&drop configuration or data file."""
        if a0 is None:
            return
        mime_data = a0.mimeData()
        if mime_data is None:
            return
        if mime_data.hasText():
            a0.accept()
        else:
            a0.ignore()

    def dropEvent(self, a0: QtGui.QDropEvent | None = None) -> None:
        """
        Set the "File" parameter when a file is dropped in the main window.

        This triggers the `load_file()` method.
        """
        if a0 is None:
            return
        mime_data = a0.mimeData()
        if mime_data is None:
            return
        for url in mime_data.urls():
            file_path = url.toLocalFile()
            if file_path.endswith(ALLOWED_FORMAT):
                self.wconfiguration.files_parameters["file"] = file_path

    def closeEvent(self, a0):
        """Quit."""
        if self.check_config_loaded():
            self.reset()
        print("Bye !")
        return super().closeEvent(a0)



"""A standalone progress bar window that emits its progress."""

from PyQt6 import QtWidgets
from PyQt6.QtCore import pyqtSlot


class PopupProgressBar(QtWidgets.QWidget):
    """Pop-up progress bar."""

    def __init__(self, min: int = 0, max: int = 100, title: str = ""):
        # Setup window
        super().__init__()
        self.pbar = QtWidgets.QProgressBar(self)
        self.pbar.setGeometry(30, 40, 500, 75)
        self.pbar.setMinimum(min)
        self.pbar.setMaximum(max)

        vlayout = QtWidgets.QVBoxLayout(self)
        vlayout.addWidget(self.pbar)
        self.setLayout(vlayout)
        self.setGeometry(300, 300, 550, 100)
        self.setWindowTitle(title)

    def start_progress(self):
        self.show()

    @pyqtSlot(int)
    def update_progress(self, idx: int):
        self.pbar.setValue(idx + 1)



"""Custom widgets for pyuson, the GUI for ultrasound experiments."""

from ._batch_processing import BatchProcessingWidget
from ._buttons import ButtonsWidget
from ._configuration import ConfigurationWidget, ParamContent
from ._graphs import GraphsWidget

__all__ = [
    "BatchProcessingWidget",
    "ButtonsWidget",
    "ConfigurationWidget",
    "GraphsWidget",
    "ParamContent",
]



"""A widget for batch-processing."""

from PyQt6 import QtWidgets
from PyQt6.QtCore import pyqtSignal

from pymagnetos.core.gui.widgets import BaseBatchProcessingWidget


class BatchProcessingWidget(BaseBatchProcessingWidget):
    """
    A batch processing widget.

    Signals
    -------
    sig_echo_index_changed : emits when the echo index changed.
    """

    sig_echo_index_changed = pyqtSignal()

    def __init__(self) -> None:
        super().__init__()
        self.add_rolling_average_checkbox()

    @property
    def echo_index(self) -> int:
        return self.spinbox_echo_index.value()

    @echo_index.setter
    def echo_index(self, value: int):
        self.spinbox_echo_index.setValue(value)

    def init_edits_widgets(self) -> QtWidgets.QWidget:
        edits_widget = super().init_edits_widgets()

        # Add the echo index spin box
        echo_index_label = QtWidgets.QLabel("Echo index", self)
        self.spinbox_echo_index = QtWidgets.QSpinBox()
        self.spinbox_echo_index.setValue(1)
        self.spinbox_echo_index.valueChanged.connect(self.sig_echo_index_changed)

        grid = QtWidgets.QHBoxLayout()
        grid.addWidget(edits_widget)
        grid.addWidget(echo_index_label)
        grid.addWidget(self.spinbox_echo_index)

        edits_widget_new = QtWidgets.QWidget(self)
        edits_widget_new.setLayout(grid)

        return edits_widget_new

    def add_rolling_average_checkbox(self):
        """Add a "Rolling average checkbox below the buttons."""
        self.checkbox_rolling_average = QtWidgets.QCheckBox("Rolling average", self)
        self.buttons_grid.addWidget(self.checkbox_rolling_average, 1, 3)
        self.wbuttons.setLayout(self.buttons_grid)

    def add_findf0_checkbox(self) -> None:
        """Add a "Find f0" checkbox below the buttons."""
        if hasattr(self, "checkbox_find_f0") and isinstance(
            self.checkbox_find_f0, QtWidgets.QCheckBox
        ):
            # Checkbox already exists
            return
        self.checkbox_find_f0 = QtWidgets.QCheckBox("Find f0", self)
        self.buttons_grid.addWidget(self.checkbox_find_f0, 1, 4)
        self.wbuttons.setLayout(self.buttons_grid)



"""
A widget for batch-processing.

It includes a 3 lists of files:
- a "available" list,
- a "todo" list,
- a "done" list.
Items can be moved across the three lists. Text fields allow to filter available files.
Those lists are DropListWidget defined in this module.
"""

import glob
import os
from pathlib import Path

from PyQt6 import QtWidgets
from PyQt6.QtCore import Qt, pyqtSignal, pyqtSlot


class DropListWidget(QtWidgets.QListWidget):
    """A list widget with drag & drop."""

    def __init__(self):
        super().__init__()

        self.setAcceptDrops(True)
        self.setDragEnabled(True)
        self.setDragDropOverwriteMode(False)
        self.setSelectionMode(
            QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection
        )
        self.setDefaultDropAction(Qt.DropAction.MoveAction)
        self.setSortingEnabled(True)

    def move_item_from_other(
        self, item: QtWidgets.QListWidgetItem, otherList: QtWidgets.QListWidget
    ):
        """Move `item` from `otherList` to this list."""
        if not self.findItems(item.text(), Qt.MatchFlag.MatchExactly):
            otherList.takeItem(otherList.indexFromItem(item).row())
            self.addItem(item)

    def add_selected_items_from_other(self, otherList: QtWidgets.QListWidget):
        """Take selected items from `otherList` list and add it to this list."""
        selectedItems = otherList.selectedItems()
        for item in selectedItems:
            self.move_item_from_other(item, otherList)

    def add_all_items_from_other(self, otherList: QtWidgets.QListWidget):
        """Take all items from `other` list and add it to this list."""
        while otherList.count() != 0:
            item = otherList.takeItem(0)
            if item is not None and (
                not self.findItems(item.text(), Qt.MatchFlag.MatchExactly)
            ):
                self.addItem(item)

    def get_list_of_items(self) -> list[str]:
        """Get the list of items as strings."""
        return [self.item(idx).text() for idx in range(self.count())]  # ty:ignore[unresolved-attribute]


class BaseBatchProcessingWidget(QtWidgets.QWidget):
    """
    List widgets that can drag & drop items between each other.

    Signals
    -------
    sig_batch_process : emits when the "Batch process" is clicked.
    """

    sig_batch_process = pyqtSignal()

    def __init__(self):
        super().__init__()

        self._current_directory = ""

        layout = QtWidgets.QGridLayout()

        self.left_list = DropListWidget()
        self.right_list = DropListWidget()
        self.done_list = DropListWidget()
        self.wbuttons = self.init_buttons()

        left_list_title = QtWidgets.QLabel("Available files", self)
        left_list_title.setAlignment(Qt.AlignmentFlag.AlignCenter)

        right_list_title = QtWidgets.QLabel("Files to process", self)
        right_list_title.setAlignment(Qt.AlignmentFlag.AlignCenter)

        done_list_title = QtWidgets.QLabel("Processed files", self)

        readd_layout = QtWidgets.QHBoxLayout()
        readd_layout.addWidget(done_list_title, alignment=Qt.AlignmentFlag.AlignLeft)
        readd_layout.addWidget(
            self.init_readd_buttons(), alignment=Qt.AlignmentFlag.AlignRight
        )

        layout.addWidget(self.init_help_text(), 0, 0, 1, 2)
        layout.addWidget(self.init_edits_widgets(), 1, 0, 1, 2)
        layout.addWidget(left_list_title, 2, 0)
        layout.addWidget(self.left_list, 3, 0, 3, 1)
        layout.addWidget(right_list_title, 2, 1)
        layout.addWidget(self.right_list, 3, 1)
        layout.addLayout(readd_layout, 4, 1)
        layout.addWidget(self.done_list, 5, 1)
        layout.addWidget(self.wbuttons, 6, 0, 1, 2)

        self.setLayout(layout)

    @property
    def prefix(self) -> str:
        return self.line_prefix.text()

    @prefix.setter
    def prefix(self, value: str):
        self.line_prefix.setText(value)

    @property
    def suffix(self) -> str:
        return self.line_suffix.text()

    @suffix.setter
    def suffix(self, value: str):
        self.line_suffix.setText(value)

    @property
    def current_directory(self) -> str:
        return self._current_directory

    @current_directory.setter
    def current_directory(self, value: str):
        self._current_directory = value
        self.list_files_in_dir()

    def init_help_text(self) -> QtWidgets.QLabel:
        """Set up the help string on top of the tab."""
        help_str = (
            "Drag & drop files from the left (right) to the right (left) to add"
            " (remove, respectively) files to the batch processor. Use the file name "
            "prefix and suffix to filter the available files. The 'Batch process' will "
            "process selected files with the same settings, including echo index."
        )

        help_widget = QtWidgets.QLabel(help_str, self)
        help_widget.setWordWrap(True)

        return help_widget

    def init_edits_widgets(self) -> QtWidgets.QWidget:
        """Set up editable fields to filter files."""
        prefix_label = QtWidgets.QLabel("Prefix", self)
        self.line_prefix = QtWidgets.QLineEdit("", self)
        self.line_prefix.textChanged.connect(self.list_files_in_dir)

        suffix_label = QtWidgets.QLabel("Suffix", self)
        self.line_suffix = QtWidgets.QLineEdit(".bin", self)  # TODO : generic default
        self.line_suffix.textChanged.connect(self.list_files_in_dir)

        self.checkbox_save_as_csv = QtWidgets.QCheckBox("Save as CSV", self)
        self.checkbox_save_as_csv.setChecked(True)

        grid = QtWidgets.QHBoxLayout()
        grid.addWidget(prefix_label)
        grid.addWidget(self.line_prefix)
        grid.addWidget(suffix_label)
        grid.addWidget(self.line_suffix)
        grid.addWidget(self.checkbox_save_as_csv)

        edits_widget = QtWidgets.QWidget(self)
        edits_widget.setLayout(grid)

        return edits_widget

    def init_buttons(self) -> QtWidgets.QWidget:
        """Set up buttons to manipulate lists."""
        self.button_add_selected = QtWidgets.QPushButton("Add selected", self)
        self.button_add_selected.clicked.connect(self.add_selected)

        self.button_add_all = QtWidgets.QPushButton("Add all", self)
        self.button_add_all.clicked.connect(self.add_all)

        self.button_clear_all = QtWidgets.QPushButton("Clear all", self)
        self.button_clear_all.clicked.connect(self.clear_all)

        self.button_run = QtWidgets.QPushButton("Batch process", self)
        self.button_run.clicked.connect(self.sig_batch_process.emit)

        self.buttons_grid = QtWidgets.QGridLayout()
        self.buttons_grid.addWidget(self.button_add_selected, 0, 0)
        self.buttons_grid.addWidget(self.button_add_all, 0, 1)
        self.buttons_grid.addWidget(self.button_clear_all, 0, 2)
        self.buttons_grid.addWidget(self.button_run, 0, 3, 1, 2)

        buttons_widget = QtWidgets.QWidget(self)
        buttons_widget.setLayout(self.buttons_grid)

        return buttons_widget

    def init_readd_buttons(self) -> QtWidgets.QWidget:
        """Set up buttons to read processed files back to the queue."""
        self.button_readd_all = QtWidgets.QPushButton("⇈", self)
        self.button_readd_all.clicked.connect(self.readd_all)

        self.button_readd_selected = QtWidgets.QPushButton("↑")
        self.button_readd_selected.clicked.connect(self.readd_selected)

        grid = QtWidgets.QHBoxLayout()
        grid.addWidget(self.button_readd_selected)
        grid.addWidget(self.button_readd_all)

        readd_widget = QtWidgets.QWidget()
        readd_widget.setLayout(grid)

        return readd_widget

    @pyqtSlot()
    def list_files_in_dir(self) -> None:
        """List files filtered with suffix and prefix."""
        directory = self.current_directory

        pattern = os.path.join(str(directory), self.prefix + "*" + self.suffix)
        new_items = [
            Path(filepath).name
            for filepath in glob.glob(pattern)
            if "-pickup" not in filepath
        ]

        self.left_list.clear()
        self.left_list.addItems(new_items)

    @pyqtSlot()
    def add_selected(self) -> None:
        """Add selected files in the left list to the right list."""
        self.right_list.add_selected_items_from_other(self.left_list)

    @pyqtSlot()
    def add_all(self) -> None:
        """Add all items in the left list to the right list."""
        self.right_list.add_all_items_from_other(self.left_list)

    @pyqtSlot()
    def readd_selected(self) -> None:
        """Add selected files in the done list to the right list."""
        self.right_list.add_selected_items_from_other(self.done_list)

    @pyqtSlot()
    def readd_all(self) -> None:
        """Add all files in the done list to the right list."""
        self.right_list.add_all_items_from_other(self.done_list)

    @pyqtSlot()
    def clear_all(self) -> None:
        """Remove all items from the right list and refresh the left list."""
        self.right_list.clear()
        self.list_files_in_dir()

    def move_to_done(self, file: str) -> None:
        """Move an item from the Files to process list to the Processed files list."""
        filename = Path(file).name
        item = self.right_list.findItems(filename, Qt.MatchFlag.MatchExactly)[0]
        self.done_list.move_item_from_other(item, self.right_list)

    def get_files_to_process(self) -> list[str]:
        """List full paths to the files to process."""
        names_list = self.right_list.get_list_of_items()
        return [os.path.join(self.current_directory, name) for name in names_list]

    def enable_buttons(self) -> None:
        """Enable all buttons."""
        self.button_add_selected.setEnabled(True)
        self.button_add_all.setEnabled(True)
        self.button_clear_all.setEnabled(True)
        self.button_readd_selected.setEnabled(True)
        self.button_readd_all.setEnabled(True)
        self.button_run.setEnabled(True)

    def disable_buttons(self) -> None:
        """Disable all buttons."""
        self.button_add_selected.setEnabled(False)
        self.button_add_all.setEnabled(False)
        self.button_clear_all.setEnabled(False)
        self.button_readd_selected.setEnabled(False)
        self.button_readd_all.setEnabled(False)
        self.button_run.setEnabled(False)



"""Application for digital lock-in."""

import importlib.metadata

from PyQt6 import QtWidgets
from PyQt6.QtCore import pyqtSignal as Signal
from PyQt6.QtCore import pyqtSlot as Slot

from pymagnetos.core.gui import BaseMainWindow
from pymagnetos.core.gui.widgets import BaseBatchProcessingWidget, FileBrowserWidget

from .widgets import ButtonsWidget, ConfigurationWidget, GraphsWidget, ParamContent
from .worker import Worker

LOG_LEVEL = "INFO"
PROGRAM_NAME = "pymagnetos"
VERSION = importlib.metadata.version(PROGRAM_NAME)


class MainWindow(BaseMainWindow):
    """A graphical user interface to perform digital lock-in."""

    sig_worker_auto_phase = Signal()
    sig_worker_export_csv = Signal(str, int, bool)

    wbuttons: ButtonsWidget
    wconfiguration: ConfigurationWidget
    wfiles: FileBrowserWidget
    wgraphs: GraphsWidget
    worker: Worker

    def __init__(self) -> None:
        # Register the widgets
        self._type_wbatch = BaseBatchProcessingWidget
        self._type_wbuttons = ButtonsWidget
        self._type_wconfiguration = ConfigurationWidget
        self._type_wfiles = FileBrowserWidget
        self._type_wgraphs = GraphsWidget
        self._param_content = ParamContent
        self._type_worker = Worker

        super().__init__()

        # Initialize window
        self.setGeometry(300, 300, 900, 450)
        self.setWindowTitle("LINX")

        # Internal flags
        self._do_load_data = False

        self.logger.info(f"Running pymagnetos.linx v{VERSION}")

    def init_buttons(self) -> None:
        """Create the main buttons."""
        super().init_buttons()

        # Connect
        self.wbuttons.sig_find_f0.connect(self.preprocess)
        self.wbuttons.sig_auto_phase.connect(self.auto_phase)
        self.wbuttons.sig_export_csv.connect(self.export_csv)

    def init_log(self) -> None:
        """Initialize logging."""
        self._init_log(PROGRAM_NAME, log_level=LOG_LEVEL)

    def load_file(self) -> None:
        """
        Load a configuration or NeXus file.

        Populate the signal and reference names settings with the ones specified in the
        "measurements" section of the configuration file.
        """
        # Disable loading data before ref_name and signal_names are properly set
        self._do_load_data = False
        self._ignore_tree_change = True
        super().load_file()

        # Populate the available ref and signal names in the tree, without updating the
        # worker
        available_names = self.worker.proc._signal_names
        ref_name = self.worker.proc.ref_name
        signal_name = self.worker.proc.signal_name
        self.wconfiguration.settings_parameters.child("ref_name").setLimits(
            available_names
        )
        self.wconfiguration.settings_parameters.child("signal_name").setLimits(
            available_names
        )
        self.wconfiguration.settings_parameters["ref_name"] = ref_name
        self.wconfiguration.settings_parameters["signal_name"] = signal_name
        self._ignore_tree_change = False
        self._do_load_data = True

        if self.worker.proc.is_nexus_file:
            # If NeXus file, data and results can be shown
            self.load_data_finished()
            self.run_analysis_finished()

        # Now, data can be loaded
        if (
            self.wconfiguration.files_parameters["autoload"]
            and not self.worker.is_data_loaded
        ):
            self.load_data()

    def connect_worker(self) -> None:
        """Connect signals to worker slots."""
        super().connect_worker()

        self.sig_worker_auto_phase.connect(self.worker.auto_phase)
        self.sig_worker_export_csv.connect(self.worker.export_csv)
        self.worker.sig_export_csv_finished.connect(self.export_csv_finished)

    def run_after_tree_change(
        self, context: str, param_name: str, value: float | int | list | bool | str
    ) -> None:
        """Run additional actions after a parameter change in the parameter tree."""
        if context != "settings":
            return

        match param_name:
            case "ref_name":
                if isinstance(value, str):
                    self.update_ref_name(value)
            case "signal_name":
                if isinstance(value, str):
                    self.update_signal_name(value)
            case "time_offset":
                if isinstance(value, (float, int)):
                    self.update_time_offset(value)
            case "f0":
                if isinstance(value, (float, int)):
                    self.update_f0(value)
            case "phase_offset":
                if isinstance(value, (float, int)):
                    self.update_phase_offset(value)
            case "auto_phase_method":
                self.auto_phase()
            case _:
                self.run_analysis()

    @Slot()
    def load_data_finished(self) -> None:
        """Plot the raw data."""
        self.plot_signal()
        self.run_analysis_finished()  # the worker ran the analysis after loading data
        super().load_data_finished()

    @Slot()
    def compute_field_finished(self) -> None:
        """Plot the new field, run demodulation."""
        super().compute_field_finished()
        self.run_analysis()

    @Slot()
    def run_analysis_finished(self) -> None:
        """Plot demodulation results."""
        # Update f0 and phase offset in case it changed
        self._do_analysis = False
        self.wconfiguration.settings_parameters["f0"] = self.worker.proc.f0
        self.wconfiguration.settings_parameters["phase_offset"] = (
            self.worker.proc.phase_offset
        )
        self._do_analysis = True
        self.plot_resistance()
        self.enable_buttons()

    @Slot()
    def auto_phase(self) -> None:
        """Find the phase offset that maximises the in-phase signal."""
        if not self.check_config_loaded():
            return
        self.disable_buttons()
        self.worker.proc.cfg.settings.auto_phase_method = (
            self.wconfiguration.settings_parameters["auto_phase_method"]
        )
        self.sig_worker_auto_phase.emit()

    @Slot()
    def export_csv(self):
        """Save results as CSV."""
        if not self.check_data_loaded():
            return

        self.disable_buttons()

        default_fname = self.worker.proc.get_csv_filename()
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Save results as...",
            default_fname,
            "Text files (*.txt, *.csv, *.tsv)",
        )

        # Get saving parameters
        downsample_factor = self.wbuttons.spinbox_downsample.value()
        averaging = self.wbuttons.checkbox_downsample.isChecked()

        if fname:
            self.sig_worker_export_csv.emit(fname, downsample_factor, averaging)
        else:
            self.logger.error("[GUI] Invalid output file name for CSV file")
            self.export_csv_finished()

    @Slot()
    def export_csv_finished(self) -> None:
        """Enable buttons."""
        self.enable_buttons()

    def plot_signal(self) -> None:
        """Plot raw signal."""
        if not self.check_data_loaded():
            self.logger.warning("[GUI] Data not loaded")

        # Plot raw signal versus time
        self.wgraphs.raw_signal.clearPlots()
        self.wgraphs.raw_signal.plot(
            self.worker.proc.get_data_raw("time_exp"),
            self.worker.proc.get_data_raw(self.worker.proc.signal_name),
            pen=self.wgraphs.pen_raw_signal,
        )

        # Plot reference signal versus time
        self.wgraphs.reference.clearPlots()
        self.wgraphs.reference.plot(
            self.worker.proc.get_data_raw("time_exp"),
            self.worker.proc.get_data_raw(self.worker.proc.ref_name),
            pen=self.wgraphs.pen_reference,
        )

    def plot_resistance(self) -> None:
        """Plot the resistance after demodulation."""
        if not self.check_resistance_computed():
            self.logger.warning("[GUI] Resistance not computed, can't plot resistance")
            return

        self.wgraphs.res_time.clearPlots()
        # Plot resistance in phase versus time
        self.wgraphs.res_time.plot(
            self.worker.proc.get_data_processed("time_exp"),
            self.worker.proc.get_data_serie("resistance_in_phase"),
            pen=self.wgraphs.pen_in_phase_bup,
            name="in phase",
        )
        # Plot resistance out phase versus time
        self.wgraphs.res_time.plot(
            self.worker.proc.get_data_processed("time_exp"),
            self.worker.proc.get_data_serie("resistance_out_phase"),
            pen=self.wgraphs.pen_out_phase_bup,
            name="out phase",
        )

        self.wgraphs.res_field.clearPlots()
        if not self.check_field_aligned():
            self.logger.warning(
                "[GUI] Magnetic field not aligned, can't plot resistance versus field"
            )
            return
        # Plot resistance in phase versus field, field down
        self.wgraphs.res_field.plot(
            self.worker.proc.get_data_processed("magfield")[self.ind_bdown],
            self.worker.proc.get_data_serie("resistance_in_phase")[self.ind_bdown],
            pen=self.wgraphs.pen_in_phase_bdown,
            name="in phase, B down",
        )
        # Plot resistance in phase versus field, field up
        self.wgraphs.res_field.plot(
            self.worker.proc.get_data_processed("magfield")[self.ind_bup],
            self.worker.proc.get_data_serie("resistance_in_phase")[self.ind_bup],
            pen=self.wgraphs.pen_in_phase_bup,
            name="out phase, B up",
        )
        # Plot resistance out phase versus field, field down
        self.wgraphs.res_field.plot(
            self.worker.proc.get_data_processed("magfield")[self.ind_bdown],
            self.worker.proc.get_data_serie("resistance_out_phase")[self.ind_bdown],
            pen=self.wgraphs.pen_out_phase_bdown,
            name="out phase, B down",
        )
        # Plot resistance out phase versus field, field up
        self.wgraphs.res_field.plot(
            self.worker.proc.get_data_processed("magfield")[self.ind_bup],
            self.worker.proc.get_data_serie("resistance_out_phase")[self.ind_bup],
            pen=self.wgraphs.pen_out_phase_bup,
            name="out phase, B up",
        )

    @Slot(str)
    def update_ref_name(self, ref_name: str) -> None:
        """Update reference name in the worker."""
        if not self.check_config_loaded():
            return
        self.worker.proc.ref_name = ref_name
        self.load_data()

    @Slot(str)
    def update_signal_name(self, signal_name: str) -> None:
        """Update signal name in the worker."""
        if not self.check_config_loaded():
            return
        self.worker.proc.signal_name = signal_name
        self.load_data()

    @Slot(float)
    def update_f0(self, f0: float) -> None:
        """Update f0 in the worker."""
        if not self.check_config_loaded():
            return
        self.worker.proc.f0 = f0
        self.run_analysis()

    @Slot(float)
    def update_phase_offset(self, phase_offset: float) -> None:
        """Update phase offset in the worker."""
        if not self.check_config_loaded():
            return
        self.worker.proc.phase_offset = phase_offset
        self.run_analysis()

    @Slot(float)
    def update_time_offset(self, time_offset: float) -> None:
        """Update time offset in the worker and re-compute magnetic field."""
        if not self.check_config_loaded():
            return
        self.worker.proc.cfg.settings.time_offset = time_offset
        self.compute_field()

    def check_demodulated(self) -> bool:
        """Check if the demodulation was performed."""
        if self.check_config_loaded():
            return self.worker.proc._check_demodulated()
        return False

    def check_resistance_computed(self) -> bool:
        """Check if the resistance was computed."""
        if self.check_config_loaded():
            return self.worker.proc._check_resistance_computed()
        return False



"""
A simple file browser with a folder selection button and a checkbox.

Files can be selected, when they are, a signal is emitted.

This module includes the QTreeView itself (FileBrowser), and the host widget that adds
the folder selection button (FileBrowserWidget).
"""

from pathlib import Path

from PyQt6 import QtWidgets
from PyQt6.QtCore import QSortFilterProxyModel, pyqtSignal, pyqtSlot
from PyQt6.QtGui import QFileSystemModel


class FileBrowser(QtWidgets.QTreeView):
    """File browser and picker."""

    def __init__(self, startDir: str | Path = "") -> None:
        super().__init__()

        self.fsModel = QFileSystemModel()
        self.fsModel.setRootPath("")

        self.proxyModel = QSortFilterProxyModel()
        self.proxyModel.setSourceModel(self.fsModel)

        self.setModel(self.proxyModel)

        start_dir = Path(startDir)
        if start_dir.is_file():
            start_dir = str(start_dir.parent)
        elif start_dir.is_dir():
            start_dir = str(start_dir)
        else:
            start_dir = ""

        index = self.fsModel.index(start_dir)
        proxyIndex = self.proxyModel.mapFromSource(index)
        self.setRootIndex(proxyIndex)

    def set_directory(self, dirname: str | Path) -> None:
        """Set the current folder in the tree view."""
        dirname = str(dirname)

        index = self.fsModel.index(dirname)
        proxyIndex = self.proxyModel.mapFromSource(index)
        self.setRootIndex(proxyIndex)

    def get_current_file(self) -> str:
        """Return the currently selected file."""
        index = self.currentIndex()
        sourceIndex = self.proxyModel.mapToSource(index)

        return self.fsModel.filePath(sourceIndex)


class FileBrowserWidget(QtWidgets.QWidget):
    """
    File browser and picker with a folder selection button and a checkbox.

    pyqtSignals
    -------
    sig_file_selected : emits when an item is double-clicked. Emits True if the file is
        a TOML file, False otherwise, and the file path.
    sig_checkbox_changed : emits when the checkbox is changed.
    """

    sig_file_selected = pyqtSignal(bool, str)
    sig_checkbox_changed = pyqtSignal()

    def __init__(self) -> None:
        super().__init__()

        # Build the push button
        self.button_open_folder = QtWidgets.QPushButton("Open Folder...", self)
        self.button_open_folder.clicked.connect(self.open_folder)

        # Build the load data checkbox
        self.checkbox_autoload_data = QtWidgets.QCheckBox(
            "Load data automatically", self
        )
        self.checkbox_autoload_data.setChecked(True)
        self.checkbox_autoload_data.stateChanged.connect(self.sig_checkbox_changed)

        # Build the file browser tree view
        self.file_browser = FileBrowser()
        self.file_browser.doubleClicked.connect(self.select_file)

        # Define layout
        grid = QtWidgets.QGridLayout()
        grid.addWidget(self.button_open_folder, 0, 0)
        grid.addWidget(self.checkbox_autoload_data, 0, 1)
        grid.addWidget(self.file_browser, 1, 0, -1, 2)

        # Create widget
        self.setLayout(grid)

    @pyqtSlot()
    def open_folder(self) -> None:
        """
        Open a folder picker.

        Callback for the "Open Folder" button in the file browser.
        """
        dirname = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            "Open folder...",
        )
        if dirname:
            self.file_browser.set_directory(dirname)

    @pyqtSlot()
    def select_file(self) -> None:
        """
        Load the file selected in the file browser tab.

        Callback for when a file is double-clicked in the file browser tab.
        """
        filepath = self.file_browser.get_current_file()

        if filepath.endswith(".toml"):
            # This is a configuration file
            self.sig_file_selected.emit(True, filepath)
        else:
            # Not really sure but let's assume it is a data file used to change the
            # experiment ID
            self.sig_file_selected.emit(False, filepath)



"""The Configuration section of the GUI."""

import numpy as np
from PyQt6.QtCore import pyqtSlot
from pyqtgraph.parametertree import Parameter

from pymagnetos.core.gui.widgets import BaseConfigurationWidget

from ._param_content import ParamContent


class ConfigurationWidget(BaseConfigurationWidget):
    """Configuration Parameter Tree for digital lock-in."""

    def __init__(self, param_content: type[ParamContent]) -> None:
        super().__init__(param_content)

        self.add_extra_parameters()

    def add_extra_parameters(self) -> None:
        """Add extra parameters used only in the GUI."""
        # Show low-pass filter time constant in seconds
        self.filter_tc = Parameter.create(
            name="filter_tc",
            type="float",
            value=self.fc_to_tc(self.settings_parameters["filter_fc"]),
            title="Filter : time constant",
            readonly=True,
            suffix="s",
            siPrefix=True,
        )
        self.host_parameters.addChild(self.filter_tc)

        # Connect filter cut-off frequency change
        self.settings_parameters.child("filter_fc").sigValueChanged.connect(
            self.update_tc
        )

    @pyqtSlot()
    def update_tc(self) -> None:
        """Set low-pass filter time constant when cut-off frequency is changed."""
        self.filter_tc.setValue(self.fc_to_tc(self.settings_parameters["filter_fc"]))

    @staticmethod
    def fc_to_tc(fc: float) -> float:
        """Convert a filter cut-off frequency to its time constant."""
        return 1 / (2 * np.pi * fc)
